# -*- coding: utf-8 -*-
# Copyright (c) 2006, Mathieu Fenniak
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


"""
Implementation of generic PDF objects (dictionary, number, string, and so on).
"""
__author__ = "Mathieu Fenniak"
__author_email__ = "biziqe@mathieu.fenniak.net"

import codecs
import decimal
import logging
import re
import sys
import warnings
from sys import version_info

from PyPDF2._security import RC4_encrypt
from PyPDF2._utils import DEPR_MSG
from PyPDF2.constants import FilterTypes as FT
from PyPDF2.constants import StreamAttributes as SA
from PyPDF2.errors import (
    STREAM_TRUNCATED_PREMATURELY,
    PdfReadError,
    PdfReadWarning,
    PdfStreamError,
)

from . import _utils, filters
from ._utils import b_, chr_, ord_, readNonWhitespace, skipOverComment, u_

if version_info < (3, 0):
    from cStringIO import StringIO

    BytesIO = StringIO
else:
    from io import BytesIO, StringIO

logger = logging.getLogger(__name__)
ObjectPrefix = b_("/<[tf(n%")
NumberSigns = b_("+-")
IndirectPattern = re.compile(b_(r"[+-]?(\d+)\s+(\d+)\s+R[^a-zA-Z]"))


def read_object(stream, pdf):
    tok = stream.read(1)
    stream.seek(-1, 1)  # reset to start
    idx = ObjectPrefix.find(tok)
    if idx == 0:
        return NameObject.read_from_stream(stream, pdf)
    elif idx == 1:
        # hexadecimal string OR dictionary
        peek = stream.read(2)
        stream.seek(-2, 1)  # reset to start

        if peek == b_("<<"):
            return DictionaryObject.read_from_stream(stream, pdf)
        else:
            return readHexStringFromStream(stream)
    elif idx == 2:
        return ArrayObject.read_from_stream(stream, pdf)
    elif idx == 3 or idx == 4:
        return BooleanObject.read_from_stream(stream)
    elif idx == 5:
        return readStringFromStream(stream)
    elif idx == 6:
        return NullObject.read_from_stream(stream)
    elif idx == 7:
        # comment
        while tok not in (b_("\r"), b_("\n")):
            tok = stream.read(1)
            # Prevents an infinite loop by raising an error if the stream is at
            # the EOF
            if len(tok) <= 0:
                raise PdfStreamError("File ended unexpectedly.")
        tok = readNonWhitespace(stream)
        stream.seek(-1, 1)
        return read_object(stream, pdf)
    else:
        # number object OR indirect reference
        peek = stream.read(20)
        stream.seek(-len(peek), 1)  # reset to start
        if IndirectPattern.match(peek) is not None:
            return IndirectObject.read_from_stream(stream, pdf)
        else:
            return NumberObject.read_from_stream(stream)


def readObject(stream, pdf):
    warnings.warn(
        "readObject will be deprecated with PyPDF2 2.0.0, use read_object instead",
        PendingDeprecationWarning,
        stacklevel=2,
    )
    return read_object(stream, pdf)


class PdfObject(object):
    def get_object(self):
        """Resolve indirect references."""
        return self

    def getObject(self):
        warnings.warn(
            "getObject will be removed in PyPDF2 2.0.0. Use get_object instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_object()


class NullObject(PdfObject):
    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("null"))

    @staticmethod
    def read_from_stream(stream):
        nulltxt = stream.read(4)
        if nulltxt != b_("null"):
            raise PdfReadError("Could not read Null object")
        return NullObject()

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def readFromStream(stream):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return NullObject.read_from_stream(stream)


class BooleanObject(PdfObject):
    def __init__(self, value):
        self.value = value

    def write_to_stream(self, stream, encryption_key):
        if self.value:
            stream.write(b_("true"))
        else:
            stream.write(b_("false"))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream):
        word = stream.read(4)
        if word == b_("true"):
            return BooleanObject(True)
        elif word == b_("fals"):
            stream.read(1)
            return BooleanObject(False)
        else:
            raise PdfReadError("Could not read Boolean object")

    @staticmethod
    def readFromStream(stream):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return BooleanObject.read_from_stream(stream)


class ArrayObject(list, PdfObject):
    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("["))
        for data in self:
            stream.write(b_(" "))
            data.write_to_stream(stream, encryption_key)
        stream.write(b_(" ]"))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream, pdf):
        arr = ArrayObject()
        tmp = stream.read(1)
        if tmp != b_("["):
            raise PdfReadError("Could not read array")
        while True:
            # skip leading whitespace
            tok = stream.read(1)
            while tok.isspace():
                tok = stream.read(1)
            stream.seek(-1, 1)
            # check for array ending
            peekahead = stream.read(1)
            if peekahead == b_("]"):
                break
            stream.seek(-1, 1)
            # read and append obj
            arr.append(read_object(stream, pdf))
        return arr

    @staticmethod
    def readFromStream(stream, pdf):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return ArrayObject.read_from_stream(stream, pdf)


class IndirectObject(PdfObject):
    def __init__(self, idnum, generation, pdf):
        self.idnum = idnum
        self.generation = generation
        self.pdf = pdf

    def get_object(self):
        return self.pdf.get_object(self).get_object()

    def __repr__(self):
        return "IndirectObject(%r, %r)" % (self.idnum, self.generation)

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, IndirectObject)
            and self.idnum == other.idnum
            and self.generation == other.generation
            and self.pdf is other.pdf
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("%s %s R" % (self.idnum, self.generation)))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream, pdf):
        idnum = b_("")
        while True:
            tok = stream.read(1)
            if not tok:
                raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)
            if tok.isspace():
                break
            idnum += tok
        generation = b_("")
        while True:
            tok = stream.read(1)
            if not tok:
                raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)
            if tok.isspace():
                if not generation:
                    continue
                break
            generation += tok
        r = readNonWhitespace(stream)
        if r != b_("R"):
            raise PdfReadError(
                "Error reading indirect object reference at byte %s"
                % _utils.hexStr(stream.tell())
            )
        return IndirectObject(int(idnum), int(generation), pdf)

    @staticmethod
    def readFromStream(stream, pdf):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return IndirectObject.read_from_stream(stream, pdf)


class FloatObject(decimal.Decimal, PdfObject):
    def __new__(cls, value="0", context=None):
        try:
            return decimal.Decimal.__new__(cls, _utils.str_(value), context)
        except Exception:
            try:
                return decimal.Decimal.__new__(cls, str(value))
            except decimal.InvalidOperation:
                # If this isn't a valid decimal (happens in malformed PDFs)
                # fallback to 0
                logger.warning("Invalid FloatObject {}".format(value))
                return decimal.Decimal.__new__(cls, "0")

    def __repr__(self):
        if self == self.to_integral():
            return str(self.quantize(decimal.Decimal(1)))
        else:
            # Standard formatting adds useless extraneous zeros.
            o = "%.5f" % self
            # Remove the zeros.
            while o and o[-1] == "0":
                o = o[:-1]
            return o

    def as_numeric(self):
        return float(b_(repr(self)))

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_(repr(self)))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)


class NumberObject(int, PdfObject):
    NumberPattern = re.compile(b_("[^+-.0-9]"))
    ByteDot = b_(".")

    def __new__(cls, value):
        val = int(value)
        try:
            return int.__new__(cls, val)
        except OverflowError:
            return int.__new__(cls, 0)

    def as_numeric(self):
        return int(b_(repr(self)))

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_(repr(self)))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream):
        num = _utils.readUntilRegex(stream, NumberObject.NumberPattern)
        if num.find(NumberObject.ByteDot) != -1:
            return FloatObject(num)
        else:
            return NumberObject(num)

    @staticmethod
    def readFromStream(stream):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return NumberObject.read_from_stream(stream)


def createStringObject(string):
    """
    Given a string (either a "str" or "unicode"), create a ByteStringObject or a
    TextStringObject to represent the string.
    """
    if isinstance(string, _utils.string_type):
        return TextStringObject(string)
    elif isinstance(string, _utils.bytes_type):
        try:
            if string.startswith(codecs.BOM_UTF16_BE):
                retval = TextStringObject(string.decode("utf-16"))
                retval.autodetect_utf16 = True
                return retval
            else:
                # This is probably a big performance hit here, but we need to
                # convert string objects into the text/unicode-aware version if
                # possible... and the only way to check if that's possible is
                # to try.  Some strings are strings, some are just byte arrays.
                retval = TextStringObject(decode_pdfdocencoding(string))
                retval.autodetect_pdfdocencoding = True
                return retval
        except UnicodeDecodeError:
            return ByteStringObject(string)
    else:
        raise TypeError("createStringObject should have str or unicode arg")


def readHexStringFromStream(stream):
    stream.read(1)
    txt = ""
    x = b_("")
    while True:
        tok = readNonWhitespace(stream)
        if not tok:
            raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)
        if tok == b_(">"):
            break
        x += tok
        if len(x) == 2:
            txt += chr(int(x, base=16))
            x = b_("")
    if len(x) == 1:
        x += b_("0")
    if len(x) == 2:
        txt += chr(int(x, base=16))
    return createStringObject(b_(txt))


def readStringFromStream(stream):
    tok = stream.read(1)
    parens = 1
    txt = b_("")
    while True:
        tok = stream.read(1)
        if not tok:
            raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)
        if tok == b_("("):
            parens += 1
        elif tok == b_(")"):
            parens -= 1
            if parens == 0:
                break
        elif tok == b_("\\"):
            tok = stream.read(1)
            escape_dict = {
                b_("n"): b_("\n"),
                b_("r"): b_("\r"),
                b_("t"): b_("\t"),
                b_("b"): b_("\b"),
                b_("f"): b_("\f"),
                b_("c"): b_(r"\c"),
                b_("("): b_("("),
                b_(")"): b_(")"),
                b_("/"): b_("/"),
                b_("\\"): b_("\\"),
                b_(" "): b_(" "),
                b_("/"): b_("/"),
                b_("%"): b_("%"),
                b_("<"): b_("<"),
                b_(">"): b_(">"),
                b_("["): b_("["),
                b_("]"): b_("]"),
                b_("#"): b_("#"),
                b_("_"): b_("_"),
                b_("&"): b_("&"),
                b_("$"): b_("$"),
            }
            try:
                tok = escape_dict[tok]
            except KeyError:
                if tok.isdigit():
                    # "The number ddd may consist of one, two, or three
                    # octal digits; high-order overflow shall be ignored.
                    # Three octal digits shall be used, with leading zeros
                    # as needed, if the next character of the string is also
                    # a digit." (PDF reference 7.3.4.2, p 16)
                    for _ in range(2):
                        ntok = stream.read(1)
                        if ntok.isdigit():
                            tok += ntok
                        else:
                            break
                    tok = b_(chr(int(tok, base=8)))
                elif tok in b_("\n\r"):
                    # This case is  hit when a backslash followed by a line
                    # break occurs.  If it's a multi-char EOL, consume the
                    # second character:
                    tok = stream.read(1)
                    if tok not in b_("\n\r"):
                        stream.seek(-1, 1)
                    # Then don't add anything to the actual string, since this
                    # line break was escaped:
                    tok = b_("")
                else:
                    msg = r"Unexpected escaped string: {}".format(tok.decode("utf8"))
                    # if.strict: PdfReadError(msg)
                    logger.warning(msg)
        txt += tok
    return createStringObject(txt)


class ByteStringObject(_utils.bytes_type, PdfObject):  # type: ignore
    """
    Represents a string object where the text encoding could not be determined.
    This occurs quite often, as the PDF spec doesn't provide an alternate way to
    represent strings -- for example, the encryption data stored in files (like
    /O) is clearly not text, but is still stored in a "String" object.
    """

    @property
    def original_bytes(self):
        """For compatibility with TextStringObject.original_bytes."""
        return self

    def write_to_stream(self, stream, encryption_key):
        bytearr = self
        if encryption_key:
            bytearr = RC4_encrypt(encryption_key, bytearr)
        stream.write(b_("<"))
        stream.write(_utils.hexencode(bytearr))
        stream.write(b_(">"))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)


class TextStringObject(_utils.string_type, PdfObject):  # type: ignore
    """
    Represents a string object that has been decoded into a real unicode string.
    If read from a PDF document, this string appeared to match the
    PDFDocEncoding, or contained a UTF-16BE BOM mark to cause UTF-16 decoding to
    occur.
    """

    autodetect_pdfdocencoding = False
    autodetect_utf16 = False

    @property
    def original_bytes(self):
        """
        It is occasionally possible that a text string object gets created where
        a byte string object was expected due to the autodetection mechanism --
        if that occurs, this "original_bytes" property can be used to
        back-calculate what the original encoded bytes were.
        """
        return self.get_original_bytes()

    def get_original_bytes(self):
        # We're a text string object, but the library is trying to get our raw
        # bytes.  This can happen if we auto-detected this string as text, but
        # we were wrong.  It's pretty common.  Return the original bytes that
        # would have been used to create this object, based upon the autodetect
        # method.
        if self.autodetect_utf16:
            return codecs.BOM_UTF16_BE + self.encode("utf-16be")
        elif self.autodetect_pdfdocencoding:
            return encode_pdfdocencoding(self)
        else:
            raise Exception("no information about original bytes")

    def write_to_stream(self, stream, encryption_key):
        # Try to write the string out as a PDFDocEncoding encoded string.  It's
        # nicer to look at in the PDF file.  Sadly, we take a performance hit
        # here for trying...
        try:
            bytearr = encode_pdfdocencoding(self)
        except UnicodeEncodeError:
            bytearr = codecs.BOM_UTF16_BE + self.encode("utf-16be")
        if encryption_key:
            bytearr = RC4_encrypt(encryption_key, bytearr)
            obj = ByteStringObject(bytearr)
            obj.write_to_stream(stream, None)
        else:
            stream.write(b_("("))
            for c in bytearr:
                if not chr_(c).isalnum() and c != b_(" "):
                    stream.write(b_("\\%03o" % ord_(c)))
                else:
                    stream.write(b_(chr_(c)))
            stream.write(b_(")"))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)


class NameObject(str, PdfObject):
    delimiterPattern = re.compile(b_(r"\s+|[\(\)<>\[\]{}/%]"))
    surfix = b_("/")

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_(self))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream, pdf):
        name = stream.read(1)
        if name != NameObject.surfix:
            raise PdfReadError("name read error")
        name += _utils.readUntilRegex(
            stream, NameObject.delimiterPattern, ignore_eof=True
        )
        try:
            try:
                ret = name.decode("utf-8")
            except (UnicodeEncodeError, UnicodeDecodeError):
                ret = name.decode("gbk")
            return NameObject(ret)
        except (UnicodeEncodeError, UnicodeDecodeError):
            # Name objects should represent irregular characters
            # with a '#' followed by the symbol's hex number
            if not pdf.strict:
                warnings.warn("Illegal character in Name Object", _utils.PdfReadWarning)
                return NameObject(name)
            else:
                raise PdfReadError("Illegal character in Name Object")

    @staticmethod
    def readFromStream(stream, pdf):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return NameObject.read_from_stream(stream, pdf)


class DictionaryObject(dict, PdfObject):
    def raw_get(self, key):
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        if not isinstance(key, PdfObject):
            raise ValueError("key must be PdfObject")
        if not isinstance(value, PdfObject):
            raise ValueError("value must be PdfObject")
        return dict.__setitem__(self, key, value)

    def setdefault(self, key, value=None):
        if not isinstance(key, PdfObject):
            raise ValueError("key must be PdfObject")
        if not isinstance(value, PdfObject):
            raise ValueError("value must be PdfObject")
        return dict.setdefault(self, key, value)

    def __getitem__(self, key):
        return dict.__getitem__(self, key).get_object()

    @property
    def xmp_metadata(self):
        """
        Retrieve XMP (Extensible Metadata Platform) data relevant to the
        this object, if available.

        Stability: Added in v1.12, will exist for all future v1.x releases.
        @return Returns a {@link #xmp.XmpInformation XmlInformation} instance
        that can be used to access XMP metadata from the document.  Can also
        return None if no metadata was found on the document root.
        """
        metadata = self.get("/Metadata", None)
        if metadata is None:
            return None
        metadata = metadata.get_object()
        from . import xmp

        if not isinstance(metadata, xmp.XmpInformation):
            metadata = xmp.XmpInformation(metadata)
            self[NameObject("/Metadata")] = metadata
        return metadata

    def getXmpMetadata(self):  # XmpInformation
        """
        .. deprecated:: 1.28.3
            Use :meth:`xmp_metadata` instead.
        """
        warnings.warn(
            "getXmpMetadata will be removed in PyPDF2 2.0.0. "
            "Use xmp_metadata instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.xmp_metadata

    @property
    def xmpMetadata(self):
        """
        .. deprecated:: 1.28.3
            Use :meth:`xmp_metadata` instead.
        """
        warnings.warn(
            "xmpMetadata will be removed in PyPDF2 2.0.0. Use xmp_metadata instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.xmp_metadata

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("<<\n"))
        for key, value in list(self.items()):
            key.write_to_stream(stream, encryption_key)
            stream.write(b_(" "))
            value.write_to_stream(stream, encryption_key)
            stream.write(b_("\n"))
        stream.write(b_(">>"))

    def writeToStream(self, stream, encryption_key):
        warnings.warn(
            "writeToStream will be removed in PyPDF2 2.0.0. "
            "Use write_to_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.write_to_stream(stream, encryption_key)

    @staticmethod
    def read_from_stream(stream, pdf):
        tmp = stream.read(2)
        if tmp != b_("<<"):
            raise PdfReadError(
                "Dictionary read error at byte %s: stream must begin with '<<'"
                % _utils.hexStr(stream.tell())
            )
        data = {}
        while True:
            tok = readNonWhitespace(stream)
            if tok == b_("\x00"):
                continue
            elif tok == b_("%"):
                stream.seek(-1, 1)
                skipOverComment(stream)
                continue
            if not tok:
                raise PdfStreamError(STREAM_TRUNCATED_PREMATURELY)

            if tok == b_(">"):
                stream.read(1)
                break
            stream.seek(-1, 1)
            key = read_object(stream, pdf)
            tok = readNonWhitespace(stream)
            stream.seek(-1, 1)
            value = read_object(stream, pdf)
            if not data.get(key):
                data[key] = value
            elif pdf.strict:
                # multiple definitions of key not permitted
                raise PdfReadError(
                    "Multiple definitions in dictionary at byte %s for key %s"
                    % (_utils.hexStr(stream.tell()), key)
                )
            else:
                warnings.warn(
                    "Multiple definitions in dictionary at byte %s for key %s"
                    % (_utils.hexStr(stream.tell()), key),
                    PdfReadWarning,
                )

        pos = stream.tell()
        s = readNonWhitespace(stream)
        if s == b_("s") and stream.read(5) == b_("tream"):
            eol = stream.read(1)
            # odd PDF file output has spaces after 'stream' keyword but before EOL.
            # patch provided by Danial Sandler
            while eol == b_(" "):
                eol = stream.read(1)
            if eol not in (b_("\n"), b_("\r")):
                raise PdfStreamError("Stream data must be followed by a newline")
            if eol == b_("\r"):
                # read \n after
                if stream.read(1) != b_("\n"):
                    stream.seek(-1, 1)
            # this is a stream object, not a dictionary
            if SA.LENGTH not in data:
                raise PdfStreamError("Stream length not defined")
            length = data[SA.LENGTH]
            if isinstance(length, IndirectObject):
                t = stream.tell()
                length = pdf.get_object(length)
                stream.seek(t, 0)
            data["__streamdata__"] = stream.read(length)
            e = readNonWhitespace(stream)
            ndstream = stream.read(8)
            if (e + ndstream) != b_("endstream"):
                # (sigh) - the odd PDF file has a length that is too long, so
                # we need to read backwards to find the "endstream" ending.
                # ReportLab (unknown version) generates files with this bug,
                # and Python users into PDF files tend to be our audience.
                # we need to do this to correct the streamdata and chop off
                # an extra character.
                pos = stream.tell()
                stream.seek(-10, 1)
                end = stream.read(9)
                if end == b_("endstream"):
                    # we found it by looking back one character further.
                    data["__streamdata__"] = data["__streamdata__"][:-1]
                else:
                    stream.seek(pos, 0)
                    raise PdfReadError(
                        "Unable to find 'endstream' marker after stream at byte %s."
                        % _utils.hexStr(stream.tell())
                    )
        else:
            stream.seek(pos, 0)
        if "__streamdata__" in data:
            return StreamObject.initializeFromDictionary(data)
        else:
            retval = DictionaryObject()
            retval.update(data)
            return retval

    @staticmethod
    def readFromStream(stream, pdf):
        warnings.warn(
            "readFromStream will be removed in PyPDF2 2.0.0. "
            "Use read_from_stream instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return DictionaryObject.read_from_stream(stream, pdf)


class TreeObject(DictionaryObject):
    def __init__(self):
        DictionaryObject.__init__(self)

    def hasChildren(self):
        return "/First" in self

    def __iter__(self):
        return self.children()

    def children(self):
        if not self.hasChildren():
            if sys.version_info >= (3, 5):  # PEP 479
                return
            else:
                raise StopIteration

        child = self["/First"]
        while True:
            yield child
            if child == self["/Last"]:
                if sys.version_info >= (3, 5):  # PEP 479
                    return
                else:
                    raise StopIteration
            child = child["/Next"]

    def addChild(self, child, pdf):
        warnings.warn(
            DEPR_MSG.format("addChild", "add_child"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.add_child(child, pdf)

    def add_child(self, child, pdf):  # PdfReader
        child_obj = child.get_object()
        child = pdf.getReference(child_obj)
        assert isinstance(child, IndirectObject)

        if "/First" not in self:
            self[NameObject("/First")] = child
            self[NameObject("/Count")] = NumberObject(0)
            prev = None
        else:
            prev = self["/Last"]

        self[NameObject("/Last")] = child
        self[NameObject("/Count")] = NumberObject(self[NameObject("/Count")] + 1)

        if prev:
            prev_ref = pdf.getReference(prev)
            assert isinstance(prev_ref, IndirectObject)
            child_obj[NameObject("/Prev")] = prev_ref
            prev[NameObject("/Next")] = child

        parent_ref = pdf.getReference(self)
        assert isinstance(parent_ref, IndirectObject)
        child_obj[NameObject("/Parent")] = parent_ref

    def removeChild(self, child):
        warnings.warn(
            DEPR_MSG.format("removeChild", "remove_child"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.remove_child(child)

    def remove_child(self, child):
        child_obj = child.get_object()

        if NameObject("/Parent") not in child_obj:
            raise ValueError("Removed child does not appear to be a tree item")
        elif child_obj[NameObject("/Parent")] != self:
            raise ValueError("Removed child is not a member of this tree")

        found = False
        prev_ref = None
        prev = None
        cur_ref = self[NameObject("/First")]
        cur = cur_ref.get_object()
        last_ref = self[NameObject("/Last")]
        last = last_ref.get_object()
        while cur is not None:
            if cur == child_obj:
                if prev is None:
                    if NameObject("/Next") in cur:
                        # Removing first tree node
                        next_ref = cur[NameObject("/Next")]
                        next = next_ref.get_object()
                        del next[NameObject("/Prev")]
                        self[NameObject("/First")] = next_ref
                        self[NameObject("/Count")] = self[NameObject("/Count")] - 1

                    else:
                        # Removing only tree node
                        assert self[NameObject("/Count")] == 1
                        del self[NameObject("/Count")]
                        del self[NameObject("/First")]
                        if NameObject("/Last") in self:
                            del self[NameObject("/Last")]
                else:
                    if NameObject("/Next") in cur:
                        # Removing middle tree node
                        next_ref = cur[NameObject("/Next")]
                        next = next_ref.get_object()
                        next[NameObject("/Prev")] = prev_ref
                        prev[NameObject("/Next")] = next_ref
                        self[NameObject("/Count")] = self[NameObject("/Count")] - 1
                    else:
                        # Removing last tree node
                        assert cur == last
                        del prev[NameObject("/Next")]
                        self[NameObject("/Last")] = prev_ref
                        self[NameObject("/Count")] = self[NameObject("/Count")] - 1
                found = True
                break

            prev_ref = cur_ref
            prev = cur
            if NameObject("/Next") in cur:
                cur_ref = cur[NameObject("/Next")]
                cur = cur_ref.get_object()
            else:
                cur_ref = None
                cur = None

        if not found:
            raise ValueError("Removal couldn't find item in tree")

        del child_obj[NameObject("/Parent")]
        if NameObject("/Next") in child_obj:
            del child_obj[NameObject("/Next")]
        if NameObject("/Prev") in child_obj:
            del child_obj[NameObject("/Prev")]

    def emptyTree(self):
        for child in self:
            child_obj = child.get_object()
            del child_obj[NameObject("/Parent")]
            if NameObject("/Next") in child_obj:
                del child_obj[NameObject("/Next")]
            if NameObject("/Prev") in child_obj:
                del child_obj[NameObject("/Prev")]

        if NameObject("/Count") in self:
            del self[NameObject("/Count")]
        if NameObject("/First") in self:
            del self[NameObject("/First")]
        if NameObject("/Last") in self:
            del self[NameObject("/Last")]


class StreamObject(DictionaryObject):
    def __init__(self):
        self._data = None
        self.decoded_self = None

    @property
    def decodedSelf(self):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.decoded_self

    @decodedSelf.setter
    def decodedSelf(self, value):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.decoded_self = value

    def write_to_stream(self, stream, encryption_key):
        self[NameObject(SA.LENGTH)] = NumberObject(len(self._data))
        DictionaryObject.write_to_stream(self, stream, encryption_key)
        del self[SA.LENGTH]
        stream.write(b_("\nstream\n"))
        data = self._data
        if encryption_key:
            data = RC4_encrypt(encryption_key, data)
        stream.write(data)
        stream.write(b_("\nendstream"))

    @staticmethod
    def initializeFromDictionary(data):
        if SA.FILTER in data:
            retval = EncodedStreamObject()
        else:
            retval = DecodedStreamObject()
        retval._data = data["__streamdata__"]
        del data["__streamdata__"]
        del data[SA.LENGTH]
        retval.update(data)
        return retval

    def flateEncode(self):
        warnings.warn(
            DEPR_MSG.format("flateEncode", "flate_encode"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.flate_encode()

    def flate_encode(self):
        if SA.FILTER in self:
            f = self[SA.FILTER]
            if isinstance(f, ArrayObject):
                f.insert(0, NameObject(FT.FLATE_DECODE))
            else:
                newf = ArrayObject()
                newf.append(NameObject("/FlateDecode"))
                newf.append(f)
                f = newf
        else:
            f = NameObject("/FlateDecode")
        retval = EncodedStreamObject()
        retval[NameObject(SA.FILTER)] = f
        retval._data = filters.FlateDecode.encode(self._data)
        return retval


class DecodedStreamObject(StreamObject):
    def get_data(self):
        return self._data

    def set_data(self, data):
        self._data = data

    def getData(self):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._data

    def setData(self, data):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.set_data(data)


class EncodedStreamObject(StreamObject):
    def __init__(self):
        self.decoded_self = None

    @property
    def decodedSelf(self):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.decoded_self

    @decodedSelf.setter
    def decodedSelf(self, value):
        warnings.warn(
            DEPR_MSG.format("decodedSelf", "decoded_self"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.decoded_self = value

    def get_data(self):
        if self.decoded_self:
            # cached version of decoded object
            return self.decoded_self.get_data()
        else:
            # create decoded object
            decoded = DecodedStreamObject()

            decoded._data = filters.decode_stream_data(self)
            for key, value in list(self.items()):
                if key not in (SA.LENGTH, SA.FILTER, SA.DECODE_PARMS):
                    decoded[key] = value
            self.decoded_self = decoded
            return decoded._data

    def set_data(self, data):
        raise PdfReadError("Creating EncodedStreamObject is not currently supported")


class ContentStream(DecodedStreamObject):
    def __init__(self, stream, pdf):
        self.pdf = pdf
        self.operations = []
        # stream may be a StreamObject or an ArrayObject containing
        # multiple StreamObjects to be cat'd together.
        stream = stream.get_object()
        if isinstance(stream, ArrayObject):
            data = b_("")
            for s in stream:
                data += b_(s.get_object().get_data())
            stream = BytesIO(b_(data))
        else:
            stream = BytesIO(b_(stream.get_data()))
        self.__parseContentStream(stream)

    def __parseContentStream(self, stream):
        # file("f:\\tmp.txt", "w").write(stream.read())
        stream.seek(0, 0)
        operands = []
        while True:
            peek = readNonWhitespace(stream)
            if peek == b_("") or ord_(peek) == 0:
                break
            stream.seek(-1, 1)
            if peek.isalpha() or peek == b_("'") or peek == b_('"'):
                operator = _utils.readUntilRegex(
                    stream, NameObject.delimiterPattern, True
                )
                if operator == b_("BI"):
                    # begin inline image - a completely different parsing
                    # mechanism is required, of course... thanks buddy...
                    assert operands == []
                    ii = self._readInlineImage(stream)
                    self.operations.append((ii, b_("INLINE IMAGE")))
                else:
                    self.operations.append((operands, operator))
                    operands = []
            elif peek == b_("%"):
                # If we encounter a comment in the content stream, we have to
                # handle it here.  Typically, readObject will handle
                # encountering a comment -- but readObject assumes that
                # following the comment must be the object we're trying to
                # read.  In this case, it could be an operator instead.
                while peek not in (b_("\r"), b_("\n")):
                    peek = stream.read(1)
            else:
                operands.append(read_object(stream, None))

    def _readInlineImage(self, stream):
        # begin reading just after the "BI" - begin image
        # first read the dictionary of settings.
        settings = DictionaryObject()
        while True:
            tok = readNonWhitespace(stream)
            stream.seek(-1, 1)
            if tok == b_("I"):
                # "ID" - begin of image data
                break
            key = read_object(stream, self.pdf)
            tok = readNonWhitespace(stream)
            stream.seek(-1, 1)
            value = read_object(stream, self.pdf)
            settings[key] = value
        # left at beginning of ID
        tmp = stream.read(3)
        assert tmp[:2] == b_("ID")
        data = BytesIO()
        # Read the inline image, while checking for EI (End Image) operator.
        while True:
            # Read 8 kB at a time and check if the chunk contains the E operator.
            buf = stream.read(8192)
            # We have reached the end of the stream, but haven't found the EI operator.
            if not buf:
                raise PdfReadError("Unexpected end of stream")
            loc = buf.find(b_("E"))

            if loc == -1:
                data.write(buf)
            else:
                # Write out everything before the E.
                data.write(buf[0:loc])

                # Seek back in the stream to read the E next.
                stream.seek(loc - len(buf), 1)
                tok = stream.read(1)
                # Check for End Image
                tok2 = stream.read(1)
                if tok2 == b_("I"):
                    # Data can contain EI, so check for the Q operator.
                    tok3 = stream.read(1)
                    info = tok + tok2
                    # We need to find whitespace between EI and Q.
                    has_q_whitespace = False
                    while tok3 in _utils.WHITESPACES:
                        has_q_whitespace = True
                        info += tok3
                        tok3 = stream.read(1)
                    if tok3 == b_("Q") and has_q_whitespace:
                        stream.seek(-1, 1)
                        break
                    else:
                        stream.seek(-1, 1)
                        data.write(info)
                else:
                    stream.seek(-1, 1)
                    data.write(tok)
        return {"settings": settings, "data": data.getvalue()}

    def _getData(self):
        newdata = BytesIO()
        for operands, operator in self.operations:
            if operator == b_("INLINE IMAGE"):
                newdata.write(b_("BI"))
                dicttext = BytesIO()
                operands["settings"].write_to_stream(dicttext, None)
                newdata.write(dicttext.getvalue()[2:-2])
                newdata.write(b_("ID "))
                newdata.write(operands["data"])
                newdata.write(b_("EI"))
            else:
                for op in operands:
                    op.write_to_stream(newdata, None)
                    newdata.write(b_(" "))
                newdata.write(b_(operator))
            newdata.write(b_("\n"))
        return newdata.getvalue()

    def _setData(self, value):
        self.__parseContentStream(BytesIO(b_(value)))

    _data = property(_getData, _setData)


class RectangleObject(ArrayObject):
    """
    This class is used to represent *page boxes* in PyPDF2. These boxes include:

        * :attr:`artbox <PyPDF2._page.PageObject.artbox>`
        * :attr:`bleedbox <PyPDF2._page.PageObject.bleedbox>`
        * :attr:`cropbox <PyPDF2._page.PageObject.cropbox>`
        * :attr:`mediabox <PyPDF2._page.PageObject.mediabox>`
        * :attr:`trimbox <PyPDF2._page.PageObject.trimbox>`
    """

    def __init__(self, arr):
        # must have four points
        assert len(arr) == 4
        # automatically convert arr[x] into NumberObject(arr[x]) if necessary
        ArrayObject.__init__(self, [self._ensure_is_number(x) for x in arr])

    def _ensure_is_number(self, value):
        if not isinstance(value, (NumberObject, FloatObject)):
            value = FloatObject(value)
        return value

    def ensureIsNumber(self, value):
        warnings.warn(
            "ensureIsNumber will be removed in PyPDF2 2.0.0. ",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._ensure_is_number(value)

    def __repr__(self):
        return "RectangleObject(%s)" % repr(list(self))

    @property
    def left(self):
        return self[0]

    @property
    def bottom(self):
        return self[1]

    @property
    def right(self):
        return self[2]

    @property
    def top(self):
        return self[3]

    def getLowerLeft_x(self):
        warnings.warn(
            DEPR_MSG.format("getLowerLeft_x", "left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.left

    def getLowerLeft_y(self):
        warnings.warn(
            DEPR_MSG.format("getLowerLeft_y", "bottom"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.bottom

    def getUpperRight_x(self):
        warnings.warn(
            DEPR_MSG.format("getUpperRight_x", "right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.right

    def getUpperRight_y(self):
        warnings.warn(
            DEPR_MSG.format("getUpperRight_y", "top"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.top

    def getUpperLeft_x(self):
        warnings.warn(
            DEPR_MSG.format("getUpperLeft_x", "left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.left

    def getUpperLeft_y(self):
        warnings.warn(
            DEPR_MSG.format("getUpperLeft_y", "top"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.top

    def getLowerRight_x(self):
        warnings.warn(
            DEPR_MSG.format("getLowerRight_x", "right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.right

    def getLowerRight_y(self):
        warnings.warn(
            DEPR_MSG.format("getLowerRight_y", "bottom"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.bottom

    @property
    def lower_left(self):
        """
        Property to read and modify the lower left coordinate of this box
        in (x,y) form.
        """
        return self.left, self.bottom

    @lower_left.setter
    def lower_left(self, value):
        self[0], self[1] = [self._ensure_is_number(x) for x in value]

    @property
    def lower_right(self):
        """
        Property to read and modify the lower right coordinate of this box
        in (x,y) form.
        """
        return self.right, self.bottom

    @lower_right.setter
    def lower_right(self, value):
        self[2], self[1] = [self._ensure_is_number(x) for x in value]

    @property
    def upper_left(self):
        """
        Property to read and modify the upper left coordinate of this box
        in (x,y) form.
        """
        return self.left, self.top

    @upper_left.setter
    def upper_left(self, value):
        self[0], self[3] = [self._ensure_is_number(x) for x in value]

    @property
    def upper_right(self):
        """
        Property to read and modify the upper right coordinate of this box
        in (x,y) form.
        """
        return self.right, self.top

    @upper_right.setter
    def upper_right(self, value):
        self[2], self[3] = [self._ensure_is_number(x) for x in value]

    def getLowerLeft(self):
        warnings.warn(
            DEPR_MSG.format("getLowerLeft", "lower_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.lower_left

    def getLowerRight(self):
        warnings.warn(
            DEPR_MSG.format("getLowerRight", "lower_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.lower_right

    def getUpperLeft(self):
        warnings.warn(
            DEPR_MSG.format("getUpperLeft", "upper_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.upper_left

    def getUpperRight(self):
        warnings.warn(
            DEPR_MSG.format("getUpperRight", "upper_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.upper_right

    def setLowerLeft(self, value):
        warnings.warn(
            DEPR_MSG.format("setLowerLeft", "lower_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.lower_left = value

    def setLowerRight(self, value):
        warnings.warn(
            DEPR_MSG.format("setLowerRight", "lower_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self[2], self[1] = [self._ensure_is_number(x) for x in value]

    def setUpperLeft(self, value):
        warnings.warn(
            DEPR_MSG.format("setUpperLeft", "upper_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self[0], self[3] = [self._ensure_is_number(x) for x in value]

    def setUpperRight(self, value):
        warnings.warn(
            DEPR_MSG.format("setUpperRight", "upper_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self[2], self[3] = [self._ensure_is_number(x) for x in value]

    @property
    def width(self):
        return self.right - self.left

    def getWidth(self):
        warnings.warn(DEPR_MSG.format("getWidth", "width"), DeprecationWarning)
        return self.width

    @property
    def height(self):
        return self.top - self.bottom

    def getHeight(self):
        warnings.warn(DEPR_MSG.format("getHeight", "height"), DeprecationWarning)
        return self.height

    @property
    def lowerLeft(self):
        warnings.warn(
            DEPR_MSG.format("lowerLeft", "lower_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.lower_left

    @lowerLeft.setter
    def lowerLeft(self, value):
        warnings.warn(
            DEPR_MSG.format("lowerLeft", "lower_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.lower_left = value

    @property
    def lowerRight(self):
        warnings.warn(
            DEPR_MSG.format("lowerRight", "lower_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.lower_right

    @lowerRight.setter
    def lowerRight(self, value):
        warnings.warn(
            DEPR_MSG.format("lowerRight", "lower_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.lower_right = value

    @property
    def upperLeft(self):
        warnings.warn(
            DEPR_MSG.format("upperLeft", "upper_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.upper_left

    @upperLeft.setter
    def upperLeft(self, value):
        warnings.warn(
            DEPR_MSG.format("upperLeft", "upper_left"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.upper_left = value

    @property
    def upperRight(self):
        warnings.warn(
            DEPR_MSG.format("upperRight", "upper_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.upper_right

    @upperRight.setter
    def upperRight(self, value):
        warnings.warn(
            DEPR_MSG.format("upperRight", "upper_right"),
            PendingDeprecationWarning,
            stacklevel=2,
        )
        self.upper_right = value


class Field(TreeObject):
    """
    A class representing a field dictionary. This class is accessed through
    :meth:`get_fields()<PyPDF2.PdfReader.get_fields>`
    """

    def __init__(self, data):
        DictionaryObject.__init__(self)
        attributes = (
            "/FT",
            "/Parent",
            "/Kids",
            "/T",
            "/TU",
            "/TM",
            "/Ff",
            "/V",
            "/DV",
            "/AA",
        )
        for attr in attributes:
            try:
                self[NameObject(attr)] = data[attr]
            except KeyError:
                pass

    @property
    def field_type(self):
        """Read-only property accessing the type of this field."""
        return self.get("/FT")

    @property
    def fieldType(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`field_type` instead.
        """
        warnings.warn(
            "fieldType will be removed in PyPDF2 2.0.0. "
            "Use the field_type property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.field_type

    @property
    def parent(self):
        """Read-only property accessing the parent of this field."""
        return self.get("/Parent")

    @property
    def kids(self):
        """Read-only property accessing the kids of this field."""
        return self.get("/Kids")

    @property
    def name(self):
        """Read-only property accessing the name of this field."""
        return self.get("/T")

    @property
    def alternate_name(self):
        """Read-only property accessing the alternate name of this field."""
        return self.get("/TU")

    @property
    def altName(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`alternate_name` instead.
        """
        warnings.warn(
            "altName will be removed in PyPDF2 2.0.0. "
            "Use the alternate_name property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.alternate_name

    @property
    def mapping_name(self):
        """
        Read-only property accessing the mapping name of this field. This
        name is used by PyPDF2 as a key in the dictionary returned by
        :meth:`get_fields()<PyPDF2.PdfReader.get_fields>`
        """
        return self.get("/TM")

    @property
    def mappingName(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`mapping_name` instead.
        """
        warnings.warn(
            "mappingName will be removed in PyPDF2 2.0.0. "
            "Use the mapping_name property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.mapping_name

    @property
    def flags(self):
        """
        Read-only property accessing the field flags, specifying various
        characteristics of the field (see Table 8.70 of the PDF 1.7 reference).
        """
        return self.get("/Ff")

    @property
    def value(self):
        """
        Read-only property accessing the value of this field. Format
        varies based on field type.
        """
        return self.get("/V")

    @property
    def default_value(self):
        """Read-only property accessing the default value of this field."""
        return self.get("/DV")

    @property
    def defaultValue(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`default_value` instead.
        """
        warnings.warn(
            "defaultValue will be removed in PyPDF2 2.0.0. "
            "Use the default_value property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.default_value

    @property
    def additional_actions(self):
        """
        Read-only property accessing the additional actions dictionary.
        This dictionary defines the field's behavior in response to trigger events.
        See Section 8.5.2 of the PDF 1.7 reference.
        """
        self.get("/AA")

    @property
    def additionalActions(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`additional_actions` instead.
        """
        warnings.warn(
            "additionalActions will be removed in PyPDF2 2.0.0. "
            "Use the additional_actions property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.additional_actions


class Destination(TreeObject):
    """
    A class representing a destination within a PDF file.
    See section 8.2.1 of the PDF 1.6 reference.

    :param str title: Title of this destination.
    :param IndirectObject page: Reference to the page of this destination. Should
        be an instance of :class:`IndirectObject<PyPDF2.generic.IndirectObject>`.
    :param str typ: How the destination is displayed.
    :param args: Additional arguments may be necessary depending on the type.
    :raises PdfReadError: If destination type is invalid.

    .. list-table:: Valid ``typ`` arguments (see PDF spec for details)
       :widths: 50 50

       * - /Fit
         - No additional arguments
       * - /XYZ
         - [left] [top] [zoomFactor]
       * - /FitH
         - [top]
       * - /FitV
         - [left]
       * - /FitR
         - [left] [bottom] [right] [top]
       * - /FitB
         - No additional arguments
       * - /FitBH
         - [top]
       * - /FitBV
         - [left]
    """

    def __init__(self, title, page, typ, *args):
        DictionaryObject.__init__(self)
        self[NameObject("/Title")] = title
        self[NameObject("/Page")] = page
        self[NameObject("/Type")] = typ

        from PyPDF2.constants import TypArguments as TA
        from PyPDF2.constants import TypFitArguments as TF

        # from table 8.2 of the PDF 1.7 reference.
        if typ == "/XYZ":
            (
                self[NameObject(TA.LEFT)],
                self[NameObject(TA.TOP)],
                self[NameObject("/Zoom")],
            ) = args
        elif typ == TF.FIT_R:
            (
                self[NameObject(TA.LEFT)],
                self[NameObject(TA.BOTTOM)],
                self[NameObject(TA.RIGHT)],
                self[NameObject(TA.TOP)],
            ) = args
        elif typ in [TF.FIT_H, TF.FIT_BH]:
            (self[NameObject(TA.TOP)],) = args
        elif typ in [TF.FIT_V, TF.FIT_BV]:
            (self[NameObject(TA.LEFT)],) = args
        elif typ in [TF.FIT, TF.FIT_B]:
            pass
        else:
            raise PdfReadError("Unknown Destination Type: %r" % typ)

    @property
    def dest_array(self):
        return ArrayObject(
            [self.raw_get("/Page"), self["/Type"]]
            + [
                self[x]
                for x in ["/Left", "/Bottom", "/Right", "/Top", "/Zoom"]
                if x in self
            ]
        )

    def getDestArray(self):
        """
        .. deprecated:: 1.28.3
            Use :py:attr:`dest_array` instead.
        """
        warnings.warn(
            "getDestArray will be removed in PyPDF2 2.0.0. "
            "Use the dest_array property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.dest_array

    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("<<\n"))
        key = NameObject("/D")
        key.write_to_stream(stream, encryption_key)
        stream.write(b_(" "))
        value = self.dest_array
        value.write_to_stream(stream, encryption_key)

        key = NameObject("/S")
        key.write_to_stream(stream, encryption_key)
        stream.write(b_(" "))
        value = NameObject("/GoTo")
        value.write_to_stream(stream, encryption_key)

        stream.write(b_("\n"))
        stream.write(b_(">>"))

    @property
    def title(self):
        """
        Read-only property accessing the destination title.

        :rtype: str
        """
        return self.get("/Title")

    @property
    def page(self):
        """
        Read-only property accessing the destination page number.

        :rtype: int
        """
        return self.get("/Page")

    @property
    def typ(self):
        """
        Read-only property accessing the destination type.

        :rtype: str
        """
        return self.get("/Type")

    @property
    def zoom(self):
        """
        Read-only property accessing the zoom factor.

        :rtype: int, or ``None`` if not available.
        """
        return self.get("/Zoom", None)

    @property
    def left(self):
        """
        Read-only property accessing the left horizontal coordinate.

        :rtype: int, or ``None`` if not available.
        """
        return self.get("/Left", None)

    @property
    def right(self):
        """
        Read-only property accessing the right horizontal coordinate.

        :rtype: int, or ``None`` if not available.
        """
        return self.get("/Right", None)

    @property
    def top(self):
        """
        Read-only property accessing the top vertical coordinate.

        :rtype: int, or ``None`` if not available.
        """
        return self.get("/Top", None)

    @property
    def bottom(self):
        """
        Read-only property accessing the bottom vertical coordinate.

        :rtype: int, or ``None`` if not available.
        """
        return self.get("/Bottom", None)


class Bookmark(Destination):
    def write_to_stream(self, stream, encryption_key):
        stream.write(b_("<<\n"))
        for key in [
            NameObject(x)
            for x in ["/Title", "/Parent", "/First", "/Last", "/Next", "/Prev"]
            if x in self
        ]:
            key.write_to_stream(stream, encryption_key)
            stream.write(b_(" "))
            value = self.raw_get(key)
            value.write_to_stream(stream, encryption_key)
            stream.write(b_("\n"))
        key = NameObject("/Dest")
        key.write_to_stream(stream, encryption_key)
        stream.write(b_(" "))
        value = self.dest_array
        value.write_to_stream(stream, encryption_key)
        stream.write(b_("\n"))
        stream.write(b_(">>"))


def encode_pdfdocencoding(unicode_string):
    retval = b_("")
    for c in unicode_string:
        try:
            retval += b_(chr(_pdfDocEncoding_rev[c]))
        except KeyError:
            raise UnicodeEncodeError(
                "pdfdocencoding", c, -1, -1, "does not exist in translation table"
            )
    return retval


def decode_pdfdocencoding(byte_array):
    retval = u_("")
    for b in byte_array:
        c = _pdfDocEncoding[ord_(b)]
        if c == u_("\u0000"):
            raise UnicodeDecodeError(
                "pdfdocencoding",
                _utils.barray(b),
                -1,
                -1,
                "does not exist in translation table",
            )
        retval += c
    return retval


# PDFDocEncoding Character Set: Table D.2 of PDF Reference 1.7
# C.1 Predefined encodings sorted by character name of another PDF reference
# Some indices have '\u0000' although they should have something else:
# 22: should be '\u0017'
_pdfDocEncoding = (
    u_("\u0000"),
    u_("\u0001"),
    u_("\u0002"),
    u_("\u0003"),
    u_("\u0004"),
    u_("\u0005"),
    u_("\u0006"),
    u_("\u0007"),  #  0 -  7
    u_("\u0008"),
    u_("\u0009"),
    u_("\u000a"),
    u_("\u000b"),
    u_("\u000c"),
    u_("\u000d"),
    u_("\u000e"),
    u_("\u000f"),  #  8 - 15
    u_("\u0010"),
    u_("\u0011"),
    u_("\u0012"),
    u_("\u0013"),
    u_("\u0014"),
    u_("\u0015"),
    u_("\u0000"),
    u_("\u0017"),  # 16 - 23
    u_("\u02d8"),
    u_("\u02c7"),
    u_("\u02c6"),
    u_("\u02d9"),
    u_("\u02dd"),
    u_("\u02db"),
    u_("\u02da"),
    u_("\u02dc"),  # 24 - 31
    u_("\u0020"),
    u_("\u0021"),
    u_("\u0022"),
    u_("\u0023"),
    u_("\u0024"),
    u_("\u0025"),
    u_("\u0026"),
    u_("\u0027"),  # 32 - 39
    u_("\u0028"),
    u_("\u0029"),
    u_("\u002a"),
    u_("\u002b"),
    u_("\u002c"),
    u_("\u002d"),
    u_("\u002e"),
    u_("\u002f"),  # 40 - 47
    u_("\u0030"),
    u_("\u0031"),
    u_("\u0032"),
    u_("\u0033"),
    u_("\u0034"),
    u_("\u0035"),
    u_("\u0036"),
    u_("\u0037"),  # 48 - 55
    u_("\u0038"),
    u_("\u0039"),
    u_("\u003a"),
    u_("\u003b"),
    u_("\u003c"),
    u_("\u003d"),
    u_("\u003e"),
    u_("\u003f"),  # 56 - 63
    u_("\u0040"),
    u_("\u0041"),
    u_("\u0042"),
    u_("\u0043"),
    u_("\u0044"),
    u_("\u0045"),
    u_("\u0046"),
    u_("\u0047"),  # 64 - 71
    u_("\u0048"),
    u_("\u0049"),
    u_("\u004a"),
    u_("\u004b"),
    u_("\u004c"),
    u_("\u004d"),
    u_("\u004e"),
    u_("\u004f"),  # 72 - 79
    u_("\u0050"),
    u_("\u0051"),
    u_("\u0052"),
    u_("\u0053"),
    u_("\u0054"),
    u_("\u0055"),
    u_("\u0056"),
    u_("\u0057"),  # 80 - 87
    u_("\u0058"),
    u_("\u0059"),
    u_("\u005a"),
    u_("\u005b"),
    u_("\u005c"),
    u_("\u005d"),
    u_("\u005e"),
    u_("\u005f"),  # 88 - 95
    u_("\u0060"),
    u_("\u0061"),
    u_("\u0062"),
    u_("\u0063"),
    u_("\u0064"),
    u_("\u0065"),
    u_("\u0066"),
    u_("\u0067"),  # 96 - 103
    u_("\u0068"),
    u_("\u0069"),
    u_("\u006a"),
    u_("\u006b"),
    u_("\u006c"),
    u_("\u006d"),
    u_("\u006e"),
    u_("\u006f"),  # 104 - 111
    u_("\u0070"),
    u_("\u0071"),
    u_("\u0072"),
    u_("\u0073"),
    u_("\u0074"),
    u_("\u0075"),
    u_("\u0076"),
    u_("\u0077"),  # 112 - 119
    u_("\u0078"),
    u_("\u0079"),
    u_("\u007a"),
    u_("\u007b"),
    u_("\u007c"),
    u_("\u007d"),
    u_("\u007e"),
    u_("\u0000"),  # 120 - 127
    u_("\u2022"),
    u_("\u2020"),
    u_("\u2021"),
    u_("\u2026"),
    u_("\u2014"),
    u_("\u2013"),
    u_("\u0192"),
    u_("\u2044"),  # 128 - 135
    u_("\u2039"),
    u_("\u203a"),
    u_("\u2212"),
    u_("\u2030"),
    u_("\u201e"),
    u_("\u201c"),
    u_("\u201d"),
    u_("\u2018"),  # 136 - 143
    u_("\u2019"),
    u_("\u201a"),
    u_("\u2122"),
    u_("\ufb01"),
    u_("\ufb02"),
    u_("\u0141"),
    u_("\u0152"),
    u_("\u0160"),  # 144 - 151
    u_("\u0178"),
    u_("\u017d"),
    u_("\u0131"),
    u_("\u0142"),
    u_("\u0153"),
    u_("\u0161"),
    u_("\u017e"),
    u_("\u0000"),  # 152 - 159
    u_("\u20ac"),
    u_("\u00a1"),
    u_("\u00a2"),
    u_("\u00a3"),
    u_("\u00a4"),
    u_("\u00a5"),
    u_("\u00a6"),
    u_("\u00a7"),  # 160 - 167
    u_("\u00a8"),
    u_("\u00a9"),
    u_("\u00aa"),
    u_("\u00ab"),
    u_("\u00ac"),
    u_("\u0000"),
    u_("\u00ae"),
    u_("\u00af"),  # 168 - 175
    u_("\u00b0"),
    u_("\u00b1"),
    u_("\u00b2"),
    u_("\u00b3"),
    u_("\u00b4"),
    u_("\u00b5"),
    u_("\u00b6"),
    u_("\u00b7"),  # 176 - 183
    u_("\u00b8"),
    u_("\u00b9"),
    u_("\u00ba"),
    u_("\u00bb"),
    u_("\u00bc"),
    u_("\u00bd"),
    u_("\u00be"),
    u_("\u00bf"),  # 184 - 191
    u_("\u00c0"),
    u_("\u00c1"),
    u_("\u00c2"),
    u_("\u00c3"),
    u_("\u00c4"),
    u_("\u00c5"),
    u_("\u00c6"),
    u_("\u00c7"),  # 192 - 199
    u_("\u00c8"),
    u_("\u00c9"),
    u_("\u00ca"),
    u_("\u00cb"),
    u_("\u00cc"),
    u_("\u00cd"),
    u_("\u00ce"),
    u_("\u00cf"),  # 200 - 207
    u_("\u00d0"),
    u_("\u00d1"),
    u_("\u00d2"),
    u_("\u00d3"),
    u_("\u00d4"),
    u_("\u00d5"),
    u_("\u00d6"),
    u_("\u00d7"),  # 208 - 215
    u_("\u00d8"),
    u_("\u00d9"),
    u_("\u00da"),
    u_("\u00db"),
    u_("\u00dc"),
    u_("\u00dd"),
    u_("\u00de"),
    u_("\u00df"),  # 216 - 223
    u_("\u00e0"),
    u_("\u00e1"),
    u_("\u00e2"),
    u_("\u00e3"),
    u_("\u00e4"),
    u_("\u00e5"),
    u_("\u00e6"),
    u_("\u00e7"),  # 224 - 231
    u_("\u00e8"),
    u_("\u00e9"),
    u_("\u00ea"),
    u_("\u00eb"),
    u_("\u00ec"),
    u_("\u00ed"),
    u_("\u00ee"),
    u_("\u00ef"),  # 232 - 239
    u_("\u00f0"),
    u_("\u00f1"),
    u_("\u00f2"),
    u_("\u00f3"),
    u_("\u00f4"),
    u_("\u00f5"),
    u_("\u00f6"),
    u_("\u00f7"),  # 240 - 247
    u_("\u00f8"),
    u_("\u00f9"),
    u_("\u00fa"),
    u_("\u00fb"),
    u_("\u00fc"),
    u_("\u00fd"),
    u_("\u00fe"),
    u_("\u00ff"),  # 248 - 255
)

assert len(_pdfDocEncoding) == 256

_pdfDocEncoding_rev = {}
for i in range(256):
    char = _pdfDocEncoding[i]
    if char == u_("\u0000"):
        continue
    assert char not in _pdfDocEncoding_rev, (
        str(char) + " at " + str(i) + " already at " + str(_pdfDocEncoding_rev[char])
    )
    _pdfDocEncoding_rev[char] = i



# -*- coding: utf-8 -*-
from io import BytesIO

import pytest

from PyPDF2.constants import TypFitArguments as TF
from PyPDF2.errors import PdfReadError, PdfStreamError
from PyPDF2.generic import (
    ArrayObject,
    Bookmark,
    BooleanObject,
    ByteStringObject,
    Destination,
    DictionaryObject,
    FloatObject,
    IndirectObject,
    NameObject,
    NullObject,
    NumberObject,
    PdfObject,
    RectangleObject,
    TextStringObject,
    createStringObject,
    encode_pdfdocencoding,
    read_object,
    readHexStringFromStream,
    readStringFromStream,
)


def test_float_object_exception():
    assert FloatObject("abc") == 0


def test_number_object_exception():
    with pytest.raises(OverflowError):
        NumberObject(1.5 * 2**10000)


def test_createStringObject_exception():
    with pytest.raises(TypeError) as exc:
        createStringObject(123)
    assert exc.value.args[0] == "createStringObject should have str or unicode arg"


@pytest.mark.parametrize(
    ("value", "expected", "tell"), [(b"true", b"true", 4), (b"false", b"false", 5)]
)
def test_boolean_object(value, expected, tell):
    stream = BytesIO(value)
    assert BooleanObject.read_from_stream(stream).value == (expected == b"true")
    stream.seek(0, 0)
    assert stream.read() == expected
    assert stream.tell() == tell


def test_boolean_object_write():
    stream = BytesIO()
    boolobj = BooleanObject(None)
    boolobj.write_to_stream(stream, encryption_key=None)
    stream.seek(0, 0)
    assert stream.read() == b"false"


def test_boolean_object_exception():
    stream = BytesIO(b"False")
    with pytest.raises(PdfReadError) as exc:
        BooleanObject.read_from_stream(stream)
    assert exc.value.args[0] == "Could not read Boolean object"


def test_array_object_exception():
    stream = BytesIO(b"False")
    with pytest.raises(PdfReadError) as exc:
        ArrayObject.read_from_stream(stream, None)
    assert exc.value.args[0] == "Could not read array"


def test_null_object_exception():
    stream = BytesIO(b"notnull")
    with pytest.raises(PdfReadError) as exc:
        NullObject.read_from_stream(stream)
    assert exc.value.args[0] == "Could not read Null object"


@pytest.mark.parametrize("value", [b"", b"False", b"foo ", b"foo  ", b"foo bar"])
def test_indirect_object_premature(value):
    stream = BytesIO(value)
    with pytest.raises(PdfStreamError) as exc:
        IndirectObject.read_from_stream(stream, None)
    assert exc.value.args[0] == "Stream has ended unexpectedly"


def test_readHexStringFromStream():
    stream = BytesIO(b"a1>")
    assert readHexStringFromStream(stream) == "\x10"


def test_readHexStringFromStream_exception():
    stream = BytesIO(b"")
    with pytest.raises(PdfStreamError) as exc:
        readHexStringFromStream(stream)
    assert exc.value.args[0] == "Stream has ended unexpectedly"


def test_readStringFromStream_exception():
    stream = BytesIO(b"x")
    with pytest.raises(PdfStreamError) as exc:
        readStringFromStream(stream)
    assert exc.value.args[0] == "Stream has ended unexpectedly"


def test_readStringFromStream_not_in_escapedict_no_digit():
    stream = BytesIO(b"x\\y")
    with pytest.raises(PdfReadError) as exc:
        readStringFromStream(stream)
    assert exc.value.args[0] == "Stream has ended unexpectedly"
    # "Unexpected escaped string: y"


def test_readStringFromStream_multichar_eol():
    stream = BytesIO(b"x\\\n )")
    assert readStringFromStream(stream) == " "


def test_readStringFromStream_multichar_eol2():
    stream = BytesIO(b"x\\\n\n)")
    assert readStringFromStream(stream) == ""


def test_readStringFromStream_excape_digit():
    stream = BytesIO(b"x\\1a )")
    assert readStringFromStream(stream) == "\x01 "


def test_NameObject():
    stream = BytesIO(b"x")
    with pytest.raises(PdfReadError) as exc:
        NameObject.read_from_stream(stream, None)
    assert exc.value.args[0] == "name read error"


def test_destination_fit_r():
    d = Destination(
        NameObject("title"),
        PdfObject(),
        NameObject(TF.FIT_R),
        FloatObject(0),
        FloatObject(0),
        FloatObject(0),
        FloatObject(0),
    )
    assert d.title == NameObject("title")
    assert d.typ == "/FitR"
    assert d.zoom is None
    assert d.left == FloatObject(0)
    assert d.right == FloatObject(0)
    assert d.top == FloatObject(0)
    assert d.bottom == FloatObject(0)
    assert list(d) == []
    d.emptyTree()


def test_destination_fit_v():
    Destination(NameObject("title"), PdfObject(), NameObject(TF.FIT_V), FloatObject(0))


def test_destination_exception():
    with pytest.raises(PdfReadError):
        Destination(NameObject("title"), PdfObject(), NameObject("foo"), FloatObject(0))


def test_bookmark_write_to_stream():
    stream = BytesIO()
    bm = Bookmark(
        NameObject("title"), NameObject(), NameObject(TF.FIT_V), FloatObject(0)
    )
    bm.write_to_stream(stream, None)
    stream.seek(0, 0)
    assert stream.read() == b"<<\n/Title title\n/Dest [  /FitV 0 ]\n>>"


@pytest.mark.no_py27()
def test_encode_pdfdocencoding_keyerror():
    with pytest.raises(UnicodeEncodeError) as exc:
        encode_pdfdocencoding("😀")
    assert exc.value.args[0] == "pdfdocencoding"


def test_readObject_comment_exception():
    stream = BytesIO(b"% foobar")
    pdf = None
    with pytest.raises(PdfStreamError) as exc:
        read_object(stream, pdf)
    assert exc.value.args[0] == "File ended unexpectedly."


def test_readObject_comment():
    stream = BytesIO(b"% foobar\n1 ")
    pdf = None
    out = read_object(stream, pdf)
    assert out == 1


@pytest.mark.no_py27()
def test_ByteStringObject():
    bo = ByteStringObject("stream", encoding="utf-8")
    stream = BytesIO(b"")
    bo.write_to_stream(stream, encryption_key="foobar")
    stream.seek(0, 0)
    assert stream.read() == b"<1cdd628b972e>"  # TODO: how can we verify this?


def test_DictionaryObject_key_is_no_pdfobject():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    with pytest.raises(ValueError) as exc:
        do["foo"] = NameObject("/GoTo")
    assert exc.value.args[0] == "key must be PdfObject"


def test_DictionaryObject_xmp_meta():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    assert do.xmp_metadata is None


def test_DictionaryObject_value_is_no_pdfobject():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    with pytest.raises(ValueError) as exc:
        do[NameObject("/S")] = "/GoTo"
    assert exc.value.args[0] == "value must be PdfObject"


def test_DictionaryObject_setdefault_key_is_no_pdfobject():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    with pytest.raises(ValueError) as exc:
        do.setdefault("foo", NameObject("/GoTo"))
    assert exc.value.args[0] == "key must be PdfObject"


def test_DictionaryObject_setdefault_value_is_no_pdfobject():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    with pytest.raises(ValueError) as exc:
        do.setdefault(NameObject("/S"), "/GoTo")
    assert exc.value.args[0] == "value must be PdfObject"


def test_DictionaryObject_setdefault_value():
    do = DictionaryObject({NameObject("/S"): NameObject("/GoTo")})
    do.setdefault(NameObject("/S"), NameObject("/GoTo"))


def test_DictionaryObject_read_from_stream():
    stream = BytesIO(b"<< /S /GoTo >>")
    pdf = None
    out = DictionaryObject.read_from_stream(stream, pdf)
    assert out.get_object() == {NameObject("/S"): NameObject("/GoTo")}


def test_DictionaryObject_read_from_stream_broken():
    stream = BytesIO(b"< /S /GoTo >>")
    pdf = None
    with pytest.raises(PdfReadError) as exc:
        DictionaryObject.read_from_stream(stream, pdf)
    assert (
        exc.value.args[0]
        == "Dictionary read error at byte 0x2: stream must begin with '<<'"
    )


def test_DictionaryObject_read_from_stream_unexpected_end():
    stream = BytesIO(b"<< \x00/S /GoTo")
    pdf = None
    with pytest.raises(PdfStreamError) as exc:
        DictionaryObject.read_from_stream(stream, pdf)
    assert exc.value.args[0] == "Stream has ended unexpectedly"


def test_DictionaryObject_read_from_stream_stream_no_newline():
    stream = BytesIO(b"<< /S /GoTo >>stream")
    pdf = None
    with pytest.raises(PdfReadError) as exc:
        DictionaryObject.read_from_stream(stream, pdf)
    assert exc.value.args[0] == "Stream data must be followed by a newline"


def test_DictionaryObject_read_from_stream_stream_no_stream_length():
    stream = BytesIO(b"<< /S /GoTo >>stream\n")
    pdf = None
    with pytest.raises(PdfReadError) as exc:
        DictionaryObject.read_from_stream(stream, pdf)
    assert exc.value.args[0] == "Stream length not defined"


def test_DictionaryObject_read_from_stream_stream_stream_missing_endstream2():
    stream = BytesIO(b"<< /S /GoTo /Length 10 >>stream\n ")
    pdf = None
    with pytest.raises(PdfReadError) as exc:
        DictionaryObject.read_from_stream(stream, pdf)
    assert (
        exc.value.args[0]
        == "Unable to find 'endstream' marker after stream at byte 0x21."
    )


def test_DictionaryObject_read_from_stream_stream_stream_valid():
    stream = BytesIO(b"<< /S /GoTo /Length 10 >>stream\nBT /F1\nendstream\n")
    pdf = None
    do = DictionaryObject.read_from_stream(stream, pdf)
    # TODO: What should happen with the stream?
    assert do == {"/S": "/GoTo"}


def test_RectangleObject():
    ro = RectangleObject((1, 2, 3, 4))
    assert ro.lower_left == (1, 2)
    assert ro.lower_right == (3, 2)
    assert ro.upper_left == (1, 4)
    assert ro.upper_right == (3, 4)

    ro.lower_left = (5, 6)
    assert ro.lower_left == (5, 6)

    ro.lower_right = (7, 8)
    assert ro.lower_right == (7, 8)

    ro.upper_left = (9, 11)
    assert ro.upper_left == (9, 11)

    ro.upper_right = (13, 17)
    assert ro.upper_right == (13, 17)


def test_TextStringObject_exc():
    tso = TextStringObject("foo")
    with pytest.raises(Exception) as exc:
        tso.get_original_bytes()
    assert exc.value.args[0] == "no information about original bytes"


def test_TextStringObject_autodetect_utf16():
    tso = TextStringObject("foo")
    tso.autodetect_utf16 = True
    assert tso.get_original_bytes() == b"\xfe\xff\x00f\x00o\x00o"



# -*- coding: utf-8 -*-
#
# Copyright (c) 2006, Mathieu Fenniak
# Copyright (c) 2007, Ashish Kulkarni <kulkarni.ashish@gmail.com>
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# * Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
# * The name of the author may not be used to endorse or promote products
# derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import re
import struct
import sys
import warnings
from hashlib import md5
from sys import version_info

from PyPDF2 import _utils
from PyPDF2._page import PageObject
from PyPDF2._security import RC4_encrypt, _alg33_1, _alg34, _alg35
from PyPDF2._utils import (
    _isString,
    _VirtualList,
    b_,
    formatWarning,
    readUntilWhitespace,
)
from PyPDF2.constants import CatalogAttributes as CA
from PyPDF2.constants import Core as CO
from PyPDF2.constants import DocumentInformationAttributes as DI
from PyPDF2.constants import EncryptionDictAttributes as ED
from PyPDF2.constants import PageAttributes as PG
from PyPDF2.constants import PagesAttributes as PA
from PyPDF2.constants import StreamAttributes as SA
from PyPDF2.constants import TrailerKeys as TK
from PyPDF2.errors import PdfReadError, PdfReadWarning, PdfStreamError
from PyPDF2.generic import (
    ArrayObject,
    BooleanObject,
    ByteStringObject,
    Destination,
    DictionaryObject,
    Field,
    IndirectObject,
    NameObject,
    NullObject,
    NumberObject,
    StreamObject,
    TextStringObject,
    createStringObject,
    read_object,
    readNonWhitespace,
)

if version_info < (3, 0):
    from cStringIO import StringIO

    BytesIO = StringIO
else:
    from io import BytesIO, StringIO


def convert_to_int(d, size):
    if size > 8:
        raise PdfReadError("invalid size in convertToInt")
    d = b_("\x00\x00\x00\x00\x00\x00\x00\x00") + b_(d)
    d = d[-8:]
    return struct.unpack(">q", d)[0]


def convertToInt(d, size):
    warnings.warn(
        "convertToInt will be removed with PyPDF2 2.0.0. "
        "Use convert_to_int instead.",
        PendingDeprecationWarning,
        stacklevel=2,
    )
    return convert_to_int(d, size)


class DocumentInformation(DictionaryObject):
    """
    A class representing the basic document metadata provided in a PDF File.
    This class is accessible through
    :meth:`.getDocumentInfo()`

    All text properties of the document metadata have
    *two* properties, eg. author and author_raw. The non-raw property will
    always return a ``TextStringObject``, making it ideal for a case where
    the metadata is being displayed. The raw property can sometimes return
    a ``ByteStringObject``, if PyPDF2 was unable to decode the string's
    text encoding; this requires additional safety in the caller and
    therefore is not as commonly accessed.
    """

    def __init__(self):
        DictionaryObject.__init__(self)

    def _get_text(self, key):
        retval = self.get(key, None)
        if isinstance(retval, TextStringObject):
            return retval
        return None

    def getText(self, key):
        """
        The text value of the specified key or None.

        .. deprecated:: 1.28.0

            Use the attributes (e.g. :py:attr:`title` / :py:attr:`author`).
        """
        warnings.warn(
            "getText will be removed in PyPDF2 2.0.0.", PendingDeprecationWarning
        )
        return self._get_text(key)

    @property
    def title(self):
        """Read-only property accessing the document's **title**.
        Returns a unicode string (``TextStringObject``) or ``None``
        if the title is not specified."""
        return (
            self._get_text(DI.TITLE) or self.get(DI.TITLE).get_object()
            if self.get(DI.TITLE)
            else None
        )

    @property
    def title_raw(self):
        """The "raw" version of title; can return a ``ByteStringObject``."""
        return self.get(DI.TITLE)

    @property
    def author(self):
        """Read-only property accessing the document's **author**.
        Returns a unicode string (``TextStringObject``) or ``None``
        if the author is not specified."""
        return self._get_text(DI.AUTHOR)

    @property
    def author_raw(self):
        """The "raw" version of author; can return a ``ByteStringObject``."""
        return self.get(DI.AUTHOR)

    @property
    def subject(self):
        """Read-only property accessing the document's **subject**.
        Returns a unicode string (``TextStringObject``) or ``None``
        if the subject is not specified."""
        return self._get_text(DI.SUBJECT)

    @property
    def subject_raw(self):
        """The "raw" version of subject; can return a ``ByteStringObject``."""
        return self.get(DI.SUBJECT)

    @property
    def creator(self):
        """Read-only property accessing the document's **creator**. If the
        document was converted to PDF from another format, this is the name of the
        application (e.g. OpenOffice) that created the original document from
        which it was converted. Returns a unicode string (``TextStringObject``)
        or ``None`` if the creator is not specified."""
        return self._get_text(DI.CREATOR)

    @property
    def creator_raw(self):
        """The "raw" version of creator; can return a ``ByteStringObject``."""
        return self.get(DI.CREATOR)

    @property
    def producer(self):
        """Read-only property accessing the document's **producer**.
        If the document was converted to PDF from another format, this is
        the name of the application (for example, OSX Quartz) that converted
        it to PDF. Returns a unicode string (``TextStringObject``)
        or ``None`` if the producer is not specified."""
        return self._get_text(DI.PRODUCER)

    @property
    def producer_raw(self):
        """The "raw" version of producer; can return a ``ByteStringObject``."""
        return self.get(DI.PRODUCER)


class PdfReader(object):
    """
    Initialize a PdfReader object.

    This operation can take some time, as the PDF stream's cross-reference
    tables are read into memory.

    :param stream: A File object or an object that supports the standard read
        and seek methods similar to a File object. Could also be a
        string representing a path to a PDF file.
    :param bool strict: Determines whether user should be warned of all
        problems and also causes some correctable problems to be fatal.
        Defaults to ``True``.
    :param warndest: Destination for logging warnings (defaults to
        ``sys.stderr``). This will be removed in PyPDF2 2.0.
    :param bool overwriteWarnings: Determines whether to override Python's
        ``warnings.py`` module with a custom implementation (defaults to
        ``True``). This will be removed in PyPDF2 2.0.0.
    """

    def __init__(
        self, stream, strict=False, warndest=None, overwriteWarnings="deprecated"
    ):
        if warndest is not None:
            warnings.warn(
                "The `warndest` argument to PdfReader will be removed in PyPDF2 2.0.0.",
                PendingDeprecationWarning,
                stacklevel=2,
            )
        if overwriteWarnings:
            if overwriteWarnings != "deprecated":
                warnings.warn(
                    "The `overwriteWarnings` argument to PdfReader will be removed in PyPDF2 2.0.0.",
                    PendingDeprecationWarning,
                    stacklevel=2,
                )
            # Have to dynamically override the default showwarning since there
            # are no public methods that specify the 'file' parameter
            def _showwarning(
                message, category, filename, lineno, file=warndest, line=None
            ):
                if file is None:
                    file = sys.stderr
                try:
                    # It is possible for sys.stderr to be defined as None, most commonly in the case that the script
                    # is being run vida pythonw.exe on Windows. In this case, just swallow the warning.
                    # See also https://docs.python.org/3/library/sys.html# sys.__stderr__
                    if file is not None:
                        file.write(
                            formatWarning(message, category, filename, lineno, line)
                        )
                except IOError:
                    pass

            warnings.showwarning = _showwarning
        self.strict = strict
        self._flattened_pages = None
        self.resolved_objects = {}
        self.xref_index = 0
        self._page_id2num = None  # map page IndirectRef number to Page Number
        if hasattr(stream, "mode") and "b" not in stream.mode:
            warnings.warn(
                "PdfReader stream/file object is not in binary mode. "
                "It may not be read correctly.",
                PdfReadWarning,
            )
        if _isString(stream):
            with open(stream, "rb") as fileobj:
                stream = BytesIO(b_(fileobj.read()))
        self.read(stream)
        self.stream = stream

        self._override_encryption = False

    @property
    def flattened_pages(self):
        return self._flattened_pages

    @property
    def flattenedPages(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`flattened_pages` instead.
        """
        warnings.warn(
            "flattenedPages will be removed in PyPDF2 2.0.0. "
            "Use flattened_pages instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.flattened_pages

    @property
    def resolvedObjects(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`resolved_objects` instead.
        """
        warnings.warn(
            "resolvedObjects will be removed in PyPDF2 2.0.0. "
            "Use resolved_objects instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.resolved_objects

    @property
    def xrefIndex(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`xref_index` instead.
        """
        warnings.warn(
            "xrefIndex will be removed in PyPDF2 2.0.0. Use xref_index instead",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.xref_index

    @property
    def metadata(self):
        """
        Retrieve the PDF file's document information dictionary, if it exists.
        Note that some PDF files use metadata streams instead of docinfo
        dictionaries, and these metadata streams will not be accessed by this
        function.

        :return: the document information of this PDF file
        :rtype: :class:`DocumentInformation<pdf.DocumentInformation>` or
            ``None`` if none exists.
        """
        if TK.INFO not in self.trailer:
            return None
        obj = self.trailer[TK.INFO]
        retval = DocumentInformation()
        retval.update(obj)  # type: ignore
        return retval

    def getDocumentInfo(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`metadata` instead.
        """
        warnings.warn(
            "The `getDocumentInfo` method of PdfReader will be replaced by the "
            "`metadata` attribute in PyPDF2 2.0.0. You can switch to the "
            "metadata attribute already.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.metadata

    @property
    def documentInfo(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`metadata` instead.
        """
        warnings.warn(
            "The `documentInfo` attribute of PdfReader will be replaced by "
            "`metadata` in PyPDF2 2.0.0. You can switch to the metadata "
            "attribute already.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.metadata

    @property
    def xmp_metadata(self):
        """
        XMP (Extensible Metadata Platform) data

        :return: a :class:`XmpInformation<xmp.XmpInformation>`
            instance that can be used to access XMP metadata from the document.
        :rtype: :class:`XmpInformation<xmp.XmpInformation>` or
            ``None`` if no metadata was found on the document root.
        """
        try:
            self._override_encryption = True
            return self.trailer[TK.ROOT].xmp_metadata
        finally:
            self._override_encryption = False

    def getXmpMetadata(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`xmp_metadata` instead.
        """
        warnings.warn(
            "The `getXmpMetadata` method of PdfReader will be replaced by the "
            "`xmp_metadata` attribute in PyPDF2 2.0.0. You can switch to the "
            "xmp_metadata attribute already.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.xmp_metadata

    @property
    def xmpMetadata(self):
        """
        .. deprecated:: 1.28.0

            Use the attribute :py:attr:`xmp_metadata` instead.
        """
        warnings.warn(
            "The `xmpMetadata` attribute of PdfReader will be replaced by the "
            "`xmp_metadata` attribute in PyPDF2 2.0.0. You can switch to the "
            "xmp_metadata attribute already.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.xmp_metadata

    def _get_num_pages(self):
        """
        Calculates the number of pages in this PDF file.

        :return: number of pages
        :rtype: int
        :raises PdfReadError: if file is encrypted and restrictions prevent
            this action.
        """

        # Flattened pages will not work on an Encrypted PDF;
        # the PDF file's page count is used in this case. Otherwise,
        # the original method (flattened page count) is used.
        if self.is_encrypted:
            try:
                self._override_encryption = True
                self.decrypt("")
                return self.trailer[TK.ROOT]["/Pages"]["/Count"]  # type: ignore
            except Exception:
                raise PdfReadError("File has not been decrypted")
            finally:
                self._override_encryption = False
        else:
            if self.flattened_pages is None:
                self._flatten()
            return len(self.flattened_pages)

    def getNumPages(self):
        """
        .. deprecated:: 1.28.0

            Use :code:`len(reader.pages)` instead.
        """
        warnings.warn(
            "The `getNumPages` method of PdfReader will be removed in PyPDF2 2.0.0. "
            "Use `len(reader.pages)` instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_num_pages()

    @property
    def numPages(self):
        """
        .. deprecated:: 1.28.0

            Use :code:`len(reader.pages)` instead.
        """
        warnings.warn(
            "The `numPages` attribute of PdfReader will be removed in PyPDF2 2.0.0. "
            "Use `len(reader.pages)` instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_num_pages()

    def _get_page(self, page_number):
        """
        Retrieves a page by number from this PDF file.

        :param int pageNumber: The page number to retrieve
            (pages begin at zero)
        :return: a :class:`PageObject<PyPDF2._page.PageObject>` instance.
        :rtype: :class:`PageObject<PyPDF2._page.PageObject>`
        """
        # ensure that we're not trying to access an encrypted PDF
        # assert not self.trailer.has_key(TK.ENCRYPT)
        if self.flattened_pages is None:
            self._flatten()
        return self.flattened_pages[page_number]

    def getPage(self, pageNumber):
        """
        .. deprecated:: 1.28.0

            Use :code:`reader.pages[page_number]` instead.
        """
        warnings.warn(
            "getPage(pageNumber) will be removed in PyPDF2 2.0.0. "
            "Use reader.pages[pageNumber] instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_page(pageNumber)

    @property
    def namedDestinations(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`named_destinations` instead.
        """
        warnings.warn(
            "namedDestinations will be removed in PyPDF2 2.0.0. "
            "Use `named_destinations` instead.",
        )
        return self.named_destinations

    @property
    def named_destinations(self):
        """
        A read-only dictionary which maps names to
        :class:`Destinations<PyPDF2.generic.Destination>`
        """
        return self._get_named_destinations()

    # A select group of relevant field attributes. For the complete list,
    # see section 8.6.2 of the PDF 1.7 reference.

    def get_fields(self, tree=None, retval=None, fileobj=None):
        """
        Extracts field data if this PDF contains interactive form fields.
        The *tree* and *retval* parameters are for recursive use.

        :param fileobj: A file object (usually a text file) to write
            a report to on all interactive form fields found.
        :return: A dictionary where each key is a field name, and each
            value is a :class:`Field<PyPDF2.generic.Field>` object. By
            default, the mapping name is used for keys.
        :rtype: dict, or ``None`` if form data could not be located.
        """
        field_attributes = {
            "/FT": "Field Type",
            PA.PARENT: "Parent",
            "/T": "Field Name",
            "/TU": "Alternate Field Name",
            "/TM": "Mapping Name",
            "/Ff": "Field Flags",
            "/V": "Value",
            "/DV": "Default Value",
        }
        if retval is None:
            retval = {}
            catalog = self.trailer[TK.ROOT]
            # get the AcroForm tree
            if "/AcroForm" in catalog:
                tree = catalog["/AcroForm"]
            else:
                return None
        if tree is None:
            return retval

        self._check_kids(tree, retval, fileobj)
        for attr in field_attributes:
            if attr in tree:
                # Tree is a field
                self._build_field(tree, retval, fileobj, field_attributes)
                break

        if "/Fields" in tree:
            fields = tree["/Fields"]
            for f in fields:
                field = f.get_object()
                self._build_field(field, retval, fileobj, field_attributes)

        return retval

    def getFields(self, tree=None, retval=None, fileobj=None):
        """
        .. deprecated:: 1.28.0

            Use :meth:`get_fields` instead.
        """
        warnings.warn(
            "The getFields method of PdfReader will be removed in PyPDF2 2.0.0. "
            "Use the get_fields() method instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_fields(tree, retval, fileobj)

    def _build_field(self, field, retval, fileobj, fieldAttributes):
        self._check_kids(field, retval, fileobj)
        try:
            key = field["/TM"]
        except KeyError:
            try:
                key = field["/T"]
            except KeyError:
                # Ignore no-name field for now
                return
        if fileobj:
            self._write_field(fileobj, field, fieldAttributes)
            fileobj.write("\n")
        retval[key] = Field(field)

    def _check_kids(self, tree, retval, fileobj):
        if PA.KIDS in tree:
            # recurse down the tree
            for kid in tree[PA.KIDS]:  # type: ignore
                self.get_fields(kid.get_object(), retval, fileobj)

    def _write_field(self, fileobj, field, field_attributes):
        order = ["/TM", "/T", "/FT", PA.PARENT, "/TU", "/Ff", "/V", "/DV"]
        for attr in order:
            attr_name = field_attributes[attr]
            try:
                if attr == "/FT":
                    # Make the field type value more clear
                    types = {
                        "/Btn": "Button",
                        "/Tx": "Text",
                        "/Ch": "Choice",
                        "/Sig": "Signature",
                    }
                    if field[attr] in types:
                        fileobj.write(attr_name + ": " + types[field[attr]] + "\n")
                elif attr == PA.PARENT:
                    # Let's just write the name of the parent
                    try:
                        name = field[PA.PARENT]["/TM"]
                    except KeyError:
                        name = field[PA.PARENT]["/T"]
                    fileobj.write(attr_name + ": " + name + "\n")
                else:
                    fileobj.write(attr_name + ": " + str(field[attr]) + "\n")
            except KeyError:
                # Field attribute is N/A or unknown, so don't write anything
                pass

    def get_form_text_fields(self):
        """Retrieves form fields from the document with textual data (inputs, dropdowns)"""
        # Retrieve document form fields
        formfields = self.get_fields()
        if formfields is None:
            return {}
        return {
            formfields[field]["/T"]: formfields[field].get("/V")
            for field in formfields
            if formfields[field].get("/FT") == "/Tx"
        }

    def getFormTextFields(self):
        """
        .. deprecated:: 1.28.0

            Use :meth:`get_form_text_fields` instead.
        """
        warnings.warn(
            "The getFormTextFields method of PdfReader will be removed in PyPDF2 2.0.0. "
            "Use the get_form_text_fields() method instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_form_text_fields()

    def _get_named_destinations(self, tree=None, retval=None):
        """
        Retrieves the named destinations present in the document.

        :return: a dictionary which maps names to
            :class:`Destinations<PyPDF2.generic.Destination>`.
        :rtype: dict
        """
        if retval is None:
            retval = {}
            catalog = self.trailer[TK.ROOT]

            # get the name tree
            if CA.DESTS in catalog:
                tree = catalog[CA.DESTS]
            elif CA.NAMES in catalog:
                names = catalog[CA.NAMES]
                if CA.DESTS in names:
                    tree = names[CA.DESTS]

        if tree is None:
            return retval

        if PA.KIDS in tree:
            # recurse down the tree
            for kid in tree[PA.KIDS]:
                self._get_named_destinations(kid.get_object(), retval)

        # TABLE 3.33 Entries in a name tree node dictionary (PDF 1.7 specs)
        if CA.NAMES in tree:
            names = tree[CA.NAMES]
            for i in range(0, len(names), 2):
                key = names[i].get_object()
                val = names[i + 1].get_object()
                if isinstance(val, DictionaryObject) and "/D" in val:
                    val = val["/D"]
                dest = self._build_destination(key, val)
                if dest is not None:
                    retval[key] = dest

        return retval

    def getNamedDestinations(self, tree=None, retval=None):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`named_destinations` instead.
        """
        warnings.warn(
            "getNamedDestinations will be removed in PyPDF2 2.0.0. "
            "Use the named_destinations property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_named_destinations(tree, retval)

    @property
    def outlines(self):
        """Read-only property."""
        return self._get_outlines()

    def _get_outlines(self, node=None, outlines=None):
        """
        Retrieve the document outline present in the document.

        :return: a nested list of :class:`Destinations<PyPDF2.generic.Destination>`.
        """
        if outlines is None:
            outlines = []
            catalog = self.trailer[TK.ROOT]

            # get the outline dictionary and named destinations
            if CO.OUTLINES in catalog:
                try:
                    lines = catalog[CO.OUTLINES]
                except PdfReadError:
                    # this occurs if the /Outlines object reference is incorrect
                    # for an example of such a file, see https://unglueit-files.s3.amazonaws.com/ebf/7552c42e9280b4476e59e77acc0bc812.pdf
                    # so continue to load the file without the Bookmarks
                    return outlines

                # TABLE 8.3 Entries in the outline dictionary
                if "/First" in lines:
                    node = lines["/First"]
            self._namedDests = self._get_named_destinations()

        if node is None:
            return outlines

        # see if there are any more outlines
        while True:
            outline = self._build_outline(node)
            if outline:
                outlines.append(outline)

            # check for sub-outlines
            if "/First" in node:
                sub_outlines = []
                self._get_outlines(node["/First"], sub_outlines)
                if sub_outlines:
                    outlines.append(sub_outlines)

            if "/Next" not in node:
                break
            node = node["/Next"]

        return outlines

    def get_outlines(self, node=None, outlines=None):
        """
        .. deprecated:: 1.28.0

            Use the property :py:attr:`outlines` instead.
        """
        warnings.warn(
            "get_outlines will be removed in PyPDF2 2.0.0. "
            "Use the property :py:attr:`outlines` instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_outlines(node, outlines)

    def getOutlines(self, node=None, outlines=None):
        """
        .. deprecated:: 1.28.0

            Use the property :py:attr:`outlines` instead.
        """
        warnings.warn(
            "getOutlines will be removed in PyPDF2 2.0.0. "
            "Use the property 'outlines' instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self._get_outlines(node, outlines)

    def _get_page_number_by_indirect(self, indirect_ref):
        """Generate _pageId2Num"""
        if self._page_id2num is None:
            id2num = {}
            for i, x in enumerate(self.pages):
                id2num[x.indirectRef.idnum] = i
            self._page_id2num = id2num

        if isinstance(indirect_ref, NullObject):
            return -1
        if isinstance(indirect_ref, int):
            idnum = indirect_ref
        else:
            idnum = indirect_ref.idnum

        ret = self._page_id2num.get(idnum, -1)
        return ret

    def get_page_number(self, page):
        """
        Retrieve page number of a given PageObject

        :param PageObject page: The page to get page number. Should be
            an instance of :class:`PageObject<PyPDF2._page.PageObject>`
        :return: the page number or -1 if page not found
        :rtype: int
        """
        indirect_ref = page.indirectRef
        ret = self._get_page_number_by_indirect(indirect_ref)
        return ret

    def getPageNumber(self, page):
        """
        .. deprecated:: 1.28.0

            Use :meth:`get_page_number` instead.
        """
        warnings.warn(
            "getPageNumber will be removed in PyPDF2 2.0.0. "
            "Use get_page_number instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_page_number(page)

    def get_destination_page_number(self, destination):
        """
        Retrieve page number of a given Destination object.

        :param Destination destination: The destination to get page number.
        :return: the page number or -1 if page not found
        :rtype: int
        """
        indirect_ref = destination.page
        ret = self._get_page_number_by_indirect(indirect_ref)
        return ret

    def getDestinationPageNumber(self, destination):
        """
        .. deprecated:: 1.28.0

            Use :meth:`get_destination_page_number` instead.
        """
        warnings.warn(
            "getDestinationPageNumber will be removed in PyPDF2 2.0.0. "
            "Use get_destination_page_number instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_destination_page_number(destination)

    def _build_destination(self, title, array):
        page, typ = array[0:2]
        array = array[2:]
        try:
            return Destination(title, page, typ, *array)
        except PdfReadError:
            warnings.warn("Unknown destination : " + title + " " + str(array))
            if self.strict:
                raise
            else:
                # create a link to first Page
                return Destination(
                    title, self._get_page(0).indirectRef, TextStringObject("/Fit")
                )

    def _build_outline(self, node):
        dest, title, outline = None, None, None

        if "/A" in node and "/Title" in node:
            # Action, section 8.5 (only type GoTo supported)
            title = node["/Title"]
            action = node["/A"]
            if action["/S"] == "/GoTo":
                dest = action["/D"]
        elif "/Dest" in node and "/Title" in node:
            # Destination, section 8.2.1
            title = node["/Title"]
            dest = node["/Dest"]

        # if destination found, then create outline
        if dest:
            if isinstance(dest, ArrayObject):
                outline = self._build_destination(title, dest)
            elif _isString(dest) and dest in self._namedDests:
                outline = self._namedDests[dest]
                outline[NameObject("/Title")] = title
            else:
                raise PdfReadError("Unexpected destination %r" % dest)
        return outline

    @property
    def pages(self):
        """Read-only property that emulates a list of :py:class:`Page<PyPDF2._page.Page>` objects."""
        return _VirtualList(self._get_num_pages, self._get_page)

    @property
    def page_layout(self):
        """
        Get the page layout.

        :return: Page layout currently being used.
        :rtype: ``str``, ``None`` if not specified

        .. list-table:: Valid ``layout`` values
           :widths: 50 200

           * - /NoLayout
             - Layout explicitly not specified
           * - /SinglePage
             - Show one page at a time
           * - /OneColumn
             - Show one column at a time
           * - /TwoColumnLeft
             - Show pages in two columns, odd-numbered pages on the left
           * - /TwoColumnRight
             - Show pages in two columns, odd-numbered pages on the right
           * - /TwoPageLeft
             - Show two pages at a time, odd-numbered pages on the left
           * - /TwoPageRight
             - Show two pages at a time, odd-numbered pages on the right
        """
        try:
            return self.trailer[TK.ROOT]["/PageLayout"]
        except KeyError:
            return None

    def getPageLayout(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`page_layout` instead.
        """
        warnings.warn(
            "getPageLayout will be removed in PyPDF2 2.0.0. "
            "Use the attribute 'page_layout' instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.page_layout

    @property
    def pageLayout(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`page_layout` instead.
        """
        warnings.warn(
            "pageLayout will be removed in PyPDF2 2.0.0. "
            "Use the attribute 'page_layout' instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.page_layout

    @property
    def page_mode(self):
        """
        Get the page mode.

        :return: Page mode currently being used.
        :rtype: ``str``, ``None`` if not specified

        .. list-table:: Valid ``mode`` values
           :widths: 50 200

           * - /UseNone
             - Do not show outlines or thumbnails panels
           * - /UseOutlines
             - Show outlines (aka bookmarks) panel
           * - /UseThumbs
             - Show page thumbnails panel
           * - /FullScreen
             - Fullscreen view
           * - /UseOC
             - Show Optional Content Group (OCG) panel
           * - /UseAttachments
             - Show attachments panel
        """
        try:
            return self.trailer[TK.ROOT]["/PageMode"]  # type: ignore
        except KeyError:
            return None

    def getPageMode(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`page_mode` instead.
        """
        warnings.warn(
            "getPageMode will be removed in PyPDF2 2.0.0. "
            "Use the attribute 'page_mode' instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.page_mode

    @property
    def pageMode(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`page_mode` instead.
        """
        warnings.warn(
            "pageMode will be removed in PyPDF2 2.0.0. "
            "Use the attribute 'page_mode' instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.page_mode

    def _flatten(self, pages=None, inherit=None, indirect_ref=None):
        inheritablePageAttributes = (
            NameObject(PG.RESOURCES),
            NameObject(PG.MEDIABOX),
            NameObject(PG.CROPBOX),
            NameObject(PG.ROTATE),
        )
        if inherit is None:
            inherit = {}
        if pages is None:
            # Fix issue 327: set flattenedPages attribute only for
            # decrypted file
            catalog = self.trailer[TK.ROOT].get_object()
            pages = catalog["/Pages"].get_object()
            self._flattened_pages = []

        t = "/Pages"
        if PA.TYPE in pages:
            t = pages[PA.TYPE]

        if t == "/Pages":
            for attr in inheritablePageAttributes:
                if attr in pages:
                    inherit[attr] = pages[attr]
            for page in pages[PA.KIDS]:
                addt = {}
                if isinstance(page, IndirectObject):
                    addt["indirect_ref"] = page
                self._flatten(page.get_object(), inherit, **addt)
        elif t == "/Page":
            for attr, value in list(inherit.items()):
                # if the page has it's own value, it does not inherit the
                # parent's value:
                if attr not in pages:
                    pages[attr] = value
            page_obj = PageObject(self, indirect_ref)
            page_obj.update(pages)
            self.flattened_pages.append(page_obj)  # type: ignore

    def _get_object_from_stream(self, indirect_reference):
        # indirect reference to object in object stream
        # read the entire object stream into memory
        stmnum, idx = self.xref_objStm[indirect_reference.idnum]
        obj_stm = IndirectObject(stmnum, 0, self).get_object()
        # This is an xref to a stream, so its type better be a stream
        assert obj_stm["/Type"] == "/ObjStm"
        # /N is the number of indirect objects in the stream
        assert idx < obj_stm["/N"]
        stream_data = BytesIO(b_(obj_stm.get_data()))
        for i in range(obj_stm["/N"]):
            readNonWhitespace(stream_data)
            stream_data.seek(-1, 1)
            objnum = NumberObject.read_from_stream(stream_data)
            readNonWhitespace(stream_data)
            stream_data.seek(-1, 1)
            offset = NumberObject.read_from_stream(stream_data)
            readNonWhitespace(stream_data)
            stream_data.seek(-1, 1)
            if objnum != indirect_reference.idnum:
                # We're only interested in one object
                continue
            if self.strict and idx != i:
                raise PdfReadError("Object is in wrong index.")
            stream_data.seek(obj_stm["/First"] + offset, 0)
            try:
                obj = read_object(stream_data, self)
            except PdfStreamError as exc:
                # Stream object cannot be read. Normally, a critical error, but
                # Adobe Reader doesn't complain, so continue (in strict mode?)
                warnings.warn(
                    "Invalid stream (index %d) within object %d %d: %s"
                    % (i, indirect_reference.idnum, indirect_reference.generation, exc),
                    PdfReadWarning,
                )

                if self.strict:
                    raise PdfReadError("Can't read object stream: %s" % exc)
                # Replace with null. Hopefully it's nothing important.
                obj = NullObject()
            return obj

        if self.strict:
            raise PdfReadError("This is a fatal error in strict mode.")
        return NullObject()

    def get_object(self, indirect_reference):
        retval = self.cache_get_indirect_object(
            indirect_reference.generation, indirect_reference.idnum
        )
        if retval is not None:
            return retval
        if (
            indirect_reference.generation == 0
            and indirect_reference.idnum in self.xref_objStm
        ):
            retval = self._get_object_from_stream(indirect_reference)
        elif (
            indirect_reference.generation in self.xref
            and indirect_reference.idnum in self.xref[indirect_reference.generation]
        ):
            start = self.xref[indirect_reference.generation][indirect_reference.idnum]
            self.stream.seek(start, 0)
            idnum, generation = self.read_object_header(self.stream)
            if idnum != indirect_reference.idnum and self.xref_index:
                # Xref table probably had bad indexes due to not being zero-indexed
                if self.strict:
                    raise PdfReadError(
                        "Expected object ID (%d %d) does not match actual (%d %d); xref table not zero-indexed."
                        % (
                            indirect_reference.idnum,
                            indirect_reference.generation,
                            idnum,
                            generation,
                        )
                    )
                else:
                    pass  # xref table is corrected in non-strict mode
            elif idnum != indirect_reference.idnum and self.strict:
                # some other problem
                raise PdfReadError(
                    "Expected object ID (%d %d) does not match actual (%d %d)."
                    % (
                        indirect_reference.idnum,
                        indirect_reference.generation,
                        idnum,
                        generation,
                    )
                )
            if self.strict:
                assert generation == indirect_reference.generation
            retval = read_object(self.stream, self)

            # override encryption is used for the /Encrypt dictionary
            if not self._override_encryption and self.is_encrypted:
                # if we don't have the encryption key:
                if not hasattr(self, "_decryption_key"):
                    raise PdfReadError("file has not been decrypted")
                # otherwise, decrypt here...
                pack1 = struct.pack("<i", indirect_reference.idnum)[:3]
                pack2 = struct.pack("<i", indirect_reference.generation)[:2]
                key = self._decryption_key + pack1 + pack2
                assert len(key) == (len(self._decryption_key) + 5)
                md5_hash = md5(key).digest()
                key = md5_hash[: min(16, len(self._decryption_key) + 5)]
                retval = self._decrypt_object(retval, key)
        else:
            warnings.warn(
                "Object %d %d not defined."
                % (indirect_reference.idnum, indirect_reference.generation),
                PdfReadWarning,
            )
            if self.strict:
                raise PdfReadError("Could not find object.")
        self.cache_indirect_object(
            indirect_reference.generation, indirect_reference.idnum, retval
        )
        return retval

    def getObject(self, indirectReference):
        """
        .. deprecated:: 1.28.0

            Use :meth:`get_object` instead.
        """
        warnings.warn(
            "getObject(indirectReference) will be removed in PyPDF2 2.0.0. "
            "Use get_object(indirect_reference) instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.get_object(indirectReference)

    def _decrypt_object(self, obj, key):
        if isinstance(obj, (ByteStringObject, TextStringObject)):
            obj = createStringObject(RC4_encrypt(key, obj.original_bytes))
        elif isinstance(obj, StreamObject):
            obj._data = RC4_encrypt(key, obj._data)
        elif isinstance(obj, DictionaryObject):
            for dictkey, value in list(obj.items()):
                obj[dictkey] = self._decrypt_object(value, key)
        elif isinstance(obj, ArrayObject):
            for i in range(len(obj)):
                obj[i] = self._decrypt_object(obj[i], key)
        return obj

    def read_object_header(self, stream):
        # Should never be necessary to read out whitespace, since the
        # cross-reference table should put us in the right spot to read the
        # object header.  In reality... some files have stupid cross reference
        # tables that are off by whitespace bytes.
        extra = False
        _utils.skipOverComment(stream)
        extra |= _utils.skipOverWhitespace(stream)
        stream.seek(-1, 1)
        idnum = readUntilWhitespace(stream)
        extra |= _utils.skipOverWhitespace(stream)
        stream.seek(-1, 1)
        generation = readUntilWhitespace(stream)
        extra |= _utils.skipOverWhitespace(stream)
        stream.seek(-1, 1)

        # although it's not used, it might still be necessary to read
        _obj = stream.read(3)  # noqa: F841

        readNonWhitespace(stream)
        stream.seek(-1, 1)
        if extra and self.strict:
            warnings.warn(
                "Superfluous whitespace found in object header %s %s"
                % (idnum, generation),
                PdfReadWarning,
            )
        return int(idnum), int(generation)

    def readObjectHeader(self, stream):
        """
        .. deprecated:: 1.28.0

            Use :meth:`read_object_header` instead.
        """
        warnings.warn(
            "readObjectHeader will be removed with PyPDF2 2.0.0. "
            "Use read_object_header instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.read_object_header(stream)

    def cache_get_indirect_object(self, generation, idnum):
        return self.resolved_objects.get((generation, idnum))

    def cacheGetIndirectObject(self, generation, idnum):
        """
        .. deprecated:: 1.28.0

            Use :meth:`cache_get_indirect_object` instead.
        """
        warnings.warn(
            "cacheGetIndirectObject will be removed with PyPDF2 2.0.0. "
            "Use cache_get_indirect_object instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.cache_get_indirect_object(generation, idnum)

    def cache_indirect_object(self, generation, idnum, obj):
        if (generation, idnum) in self.resolved_objects:
            msg = "Overwriting cache for %s %s" % (generation, idnum)
            if self.strict:
                raise PdfReadError(msg)
            else:
                warnings.warn(msg)
        self.resolved_objects[(generation, idnum)] = obj
        return obj

    def cacheIndirectObject(self, generation, idnum, obj):
        """
        .. deprecated:: 1.28.0

            Use :meth:`cache_indirect_object` instead.
        """
        warnings.warn(
            "cacheIndirectObject will be removed with PyPDF2 2.0.0. "
            "Use cache_indirect_object instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.cache_indirect_object(generation, idnum, obj)

    def read(self, stream):
        # start at the end:
        stream.seek(-1, 2)
        if not stream.tell():
            raise PdfReadError("Cannot read an empty file")
        if self.strict:
            stream.seek(0, 0)
            header_byte = stream.read(5)
            if header_byte != b"%PDF-":
                raise PdfReadError(
                    "PDF starts with '{}', but '%PDF-' expected".format(
                        header_byte.decode("utf8")
                    )
                )
            stream.seek(-1, 2)
        last1M = stream.tell() - 1024 * 1024 + 1  # offset of last MB of stream
        line = b_("")
        while line[:5] != b_("%%EOF"):
            if stream.tell() < last1M:
                raise PdfReadError("EOF marker not found")
            line = self.read_next_end_line(stream)

        startxref = self._find_startxref_pos(stream)

        # check and eventually correct the startxref only in not strict
        xref_issue_nr = self._get_xref_issues(stream, startxref)
        if xref_issue_nr != 0:
            if self.strict and xref_issue_nr:
                raise PdfReadError("Broken xref table")
            else:
                warnings.warn(
                    "incorrect startxref pointer({})".format(xref_issue_nr),
                    PdfReadWarning,
                )

        # read all cross reference tables and their trailers
        self.xref = {}
        self.xref_objStm = {}
        self.trailer = DictionaryObject()
        while True:
            # load the xref table
            stream.seek(startxref, 0)
            x = stream.read(1)
            if x == b_("x"):
                self._read_standard_xref_table(stream)
                readNonWhitespace(stream)
                stream.seek(-1, 1)
                new_trailer = read_object(stream, self)
                for key, value in list(new_trailer.items()):
                    if key not in self.trailer:
                        self.trailer[key] = value
                if "/Prev" in new_trailer:
                    startxref = new_trailer["/Prev"]
                else:
                    break
            elif xref_issue_nr:
                try:
                    self._rebuild_xref_table(stream)
                    break
                except Exception:
                    xref_issue_nr = 0
            elif x.isdigit():
                xrefstream = self._read_pdf15_xref_stream(stream)

                trailer_keys = TK.ROOT, TK.ENCRYPT, TK.INFO, TK.ID
                for key in trailer_keys:
                    if key in xrefstream and key not in self.trailer:
                        self.trailer[NameObject(key)] = xrefstream.raw_get(key)
                if "/Prev" in xrefstream:
                    startxref = xrefstream["/Prev"]
                else:
                    break
            else:
                # some PDFs have /Prev=0 in the trailer, instead of no /Prev
                if startxref == 0:
                    if self.strict:
                        raise PdfReadError(
                            "/Prev=0 in the trailer (try opening with strict=False)"
                        )
                    else:
                        warnings.warn(
                            "/Prev=0 in the trailer - assuming there"
                            " is no previous xref table"
                        )
                        break
                # bad xref character at startxref.  Let's see if we can find
                # the xref table nearby, as we've observed this error with an
                # off-by-one before.
                stream.seek(-11, 1)
                tmp = stream.read(20)
                xref_loc = tmp.find(b_("xref"))
                if xref_loc != -1:
                    startxref -= 10 - xref_loc
                    continue
                # No explicit xref table, try finding a cross-reference stream.
                stream.seek(startxref, 0)
                found = False
                for look in range(5):
                    if stream.read(1).isdigit():
                        # This is not a standard PDF, consider adding a warning
                        startxref += look
                        found = True
                        break
                if found:
                    continue
                # no xref table found at specified location
                raise PdfReadError("Could not find xref table at specified location")
        # if not zero-indexed, verify that the table is correct; change it if necessary
        if self.xref_index and not self.strict:
            loc = stream.tell()
            for gen in self.xref:
                if gen == 65535:
                    continue
                for id in self.xref[gen]:
                    stream.seek(self.xref[gen][id], 0)
                    try:
                        pid, pgen = self.read_object_header(stream)
                    except ValueError:
                        break
                    if pid == id - self.xref_index:
                        self._zero_xref(gen)
                        break
                    # if not, then either it's just plain wrong, or the
                    # non-zero-index is actually correct
            stream.seek(loc, 0)  # return to where it was

    def _find_startxref_pos(self, stream):
        """Find startxref entry - the location of the xref table"""
        line = self.read_next_end_line(stream)
        try:
            startxref = int(line)
        except ValueError:
            # 'startxref' may be on the same line as the location
            if not line.startswith(b_("startxref")):
                raise PdfReadError("startxref not found")
            startxref = int(line[9:].strip())
            warnings.warn("startxref on same line as offset")
        else:
            line = self.read_next_end_line(stream)
            if line[:9] != b_("startxref"):
                raise PdfReadError("startxref not found")
        return startxref

    def _read_standard_xref_table(self, stream):
        # standard cross-reference table
        ref = stream.read(4)
        if ref[:3] != b_("ref"):
            raise PdfReadError("xref table read error")
        readNonWhitespace(stream)
        stream.seek(-1, 1)
        firsttime = True  # check if the first time looking at the xref table
        while True:
            num = read_object(stream, self)
            if firsttime and num != 0:
                self.xref_index = num
                if self.strict:
                    warnings.warn(
                        "Xref table not zero-indexed. ID numbers for objects will be corrected.",
                        PdfReadWarning,
                    )
                    # if table not zero indexed, could be due to error from when PDF was created
                    # which will lead to mismatched indices later on, only warned and corrected if self.strict=True
            firsttime = False
            readNonWhitespace(stream)
            stream.seek(-1, 1)
            size = read_object(stream, self)
            readNonWhitespace(stream)
            stream.seek(-1, 1)
            cnt = 0
            while cnt < size:
                line = stream.read(20)

                # It's very clear in section 3.4.3 of the PDF spec
                # that all cross-reference table lines are a fixed
                # 20 bytes (as of PDF 1.7). However, some files have
                # 21-byte entries (or more) due to the use of \r\n
                # (CRLF) EOL's. Detect that case, and adjust the line
                # until it does not begin with a \r (CR) or \n (LF).
                while line[0] in b_("\x0D\x0A"):
                    stream.seek(-20 + 1, 1)
                    line = stream.read(20)

                # On the other hand, some malformed PDF files
                # use a single character EOL without a preceeding
                # space.  Detect that case, and seek the stream
                # back one character.  (0-9 means we've bled into
                # the next xref entry, t means we've bled into the
                # text "trailer"):
                if line[-1] in b_("0123456789t"):
                    stream.seek(-1, 1)

                offset, generation = line[:16].split(b_(" "))
                offset, generation = int(offset), int(generation)
                if generation not in self.xref:
                    self.xref[generation] = {}
                if num in self.xref[generation]:
                    # It really seems like we should allow the last
                    # xref table in the file to override previous
                    # ones. Since we read the file backwards, assume
                    # any existing key is already set correctly.
                    pass
                else:
                    self.xref[generation][num] = offset
                cnt += 1
                num += 1
            readNonWhitespace(stream)
            stream.seek(-1, 1)
            trailertag = stream.read(7)
            if trailertag != b_("trailer"):
                # more xrefs!
                stream.seek(-7, 1)
            else:
                break

    def _read_pdf15_xref_stream(self, stream):
        # PDF 1.5+ Cross-Reference Stream
        stream.seek(-1, 1)
        idnum, generation = self.read_object_header(stream)
        xrefstream = read_object(stream, self)
        assert xrefstream["/Type"] == "/XRef"
        self.cache_indirect_object(generation, idnum, xrefstream)
        stream_data = BytesIO(b_(xrefstream.get_data()))
        # Index pairs specify the subsections in the dictionary. If
        # none create one subsection that spans everything.
        idx_pairs = xrefstream.get("/Index", [0, xrefstream.get("/Size")])
        entry_sizes = xrefstream.get("/W")
        assert len(entry_sizes) >= 3
        if self.strict and len(entry_sizes) > 3:
            raise PdfReadError("Too many entry sizes: %s" % entry_sizes)

        def get_entry(i):
            # Reads the correct number of bytes for each entry. See the
            # discussion of the W parameter in PDF spec table 17.
            if entry_sizes[i] > 0:
                d = stream_data.read(entry_sizes[i])
                return convert_to_int(d, entry_sizes[i])

            # PDF Spec Table 17: A value of zero for an element in the
            # W array indicates...the default value shall be used
            if i == 0:
                return 1  # First value defaults to 1
            else:
                return 0

        def used_before(num, generation):
            # We move backwards through the xrefs, don't replace any.
            return num in self.xref.get(generation, []) or num in self.xref_objStm

        # Iterate through each subsection
        self._read_xref_subsections(idx_pairs, get_entry, used_before)
        return xrefstream

    @staticmethod
    def _get_xref_issues(stream, startxref):
        """Return an int which indicates an issue. 0 means there is no issue."""
        stream.seek(startxref - 1, 0)  # -1 to check character before
        line = stream.read(1)
        if line not in b_("\r\n \t"):
            return 1
        line = stream.read(4)
        if line != b_("xref"):
            # not an xref so check if it is an XREF object
            line = b_("")
            while line in b_("0123456789 \t"):
                line = stream.read(1)
                if line == b_(""):
                    return 2
            line += stream.read(2)  # 1 char already read, +2 to check "obj"
            if line.lower() != b_("obj"):
                return 3
            # while stream.read(1) in b_(" \t\r\n"):
            #     pass
            # line = stream.read(256)  # check that it is xref obj
            # if b_("/xref") not in line.lower():
            #     return 4
        return 0

    def _rebuild_xref_table(self, stream):
        self.xref = {}
        stream.seek(0, 0)
        f_ = stream.read(-1)

        for m in re.finditer(b_(r"[\r\n \t][ \t]*(\d+)[ \t]+(\d+)[ \t]+obj"), f_):
            idnum = int(m.group(1))
            generation = int(m.group(2))
            if generation not in self.xref:
                self.xref[generation] = {}
            self.xref[generation][idnum] = m.start(1)
        trailer_pos = f_.rfind(b"trailer") - len(f_) + 7
        stream.seek(trailer_pos, 2)
        # code below duplicated
        readNonWhitespace(stream)
        stream.seek(-1, 1)

        # there might be something that is not a dict (see #856)
        new_trailer = read_object(stream, self)

        for key, value in list(new_trailer.items()):
            if key not in self.trailer:
                self.trailer[key] = value

    def _read_xref_subsections(self, idx_pairs, get_entry, used_before):
        last_end = 0
        for start, size in self._pairs(idx_pairs):
            # The subsections must increase
            assert start >= last_end
            last_end = start + size
            for num in range(start, start + size):
                # The first entry is the type
                xref_type = get_entry(0)
                # The rest of the elements depend on the xref_type
                if xref_type == 0:
                    # linked list of free objects
                    next_free_object = get_entry(1)  # noqa: F841
                    next_generation = get_entry(2)  # noqa: F841
                elif xref_type == 1:
                    # objects that are in use but are not compressed
                    byte_offset = get_entry(1)
                    generation = get_entry(2)
                    if generation not in self.xref:
                        self.xref[generation] = {}
                    if not used_before(num, generation):
                        self.xref[generation][num] = byte_offset
                elif xref_type == 2:
                    # compressed objects
                    objstr_num = get_entry(1)
                    obstr_idx = get_entry(2)
                    generation = 0  # PDF spec table 18, generation is 0
                    if not used_before(num, generation):
                        self.xref_objStm[num] = (objstr_num, obstr_idx)
                elif self.strict:
                    raise PdfReadError("Unknown xref type: %s" % xref_type)

    def _zero_xref(self, generation):
        self.xref[generation] = {
            k - self.xref_index: v for (k, v) in list(self.xref[generation].items())
        }

    def _pairs(self, array):
        i = 0
        while True:
            yield array[i], array[i + 1]
            i += 2
            if (i + 1) >= len(array):
                break

    def read_next_end_line(self, stream, limit_offset=0):
        line_parts = []
        while True:
            # Prevent infinite loops in malformed PDFs
            if stream.tell() == 0 or stream.tell() == limit_offset:
                raise PdfReadError("Could not read malformed PDF file")
            x = stream.read(1)
            if stream.tell() < 2:
                raise PdfReadError("EOL marker not found")
            stream.seek(-2, 1)
            if x == b_("\n") or x == b_("\r"):  ## \n = LF; \r = CR
                crlf = False
                while x == b_("\n") or x == b_("\r"):
                    x = stream.read(1)
                    if x == b_("\n") or x == b_("\r"):  # account for CR+LF
                        stream.seek(-1, 1)
                        crlf = True
                    if stream.tell() < 2:
                        raise PdfReadError("EOL marker not found")
                    stream.seek(-2, 1)
                stream.seek(
                    2 if crlf else 1, 1
                )  # if using CR+LF, go back 2 bytes, else 1
                break
            else:
                line_parts.append(x)
        line_parts.reverse()
        return b"".join(line_parts)

    def readNextEndLine(self, stream, limit_offset=0):
        """
        .. deprecated:: 1.28.0

            Use :meth:`read_next_end_line` instead.
        """
        warnings.warn(
            "readNextEndLine will be removed in PyPDF2 2.0.0. "
            "Use read_next_end_line instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.read_next_end_line(stream, limit_offset)

    def decrypt(self, password):
        """
        When using an encrypted / secured PDF file with the PDF Standard
        encryption handler, this function will allow the file to be decrypted.
        It checks the given password against the document's user password and
        owner password, and then stores the resulting decryption key if either
        password is correct.

        It does not matter which password was matched.  Both passwords provide
        the correct decryption key that will allow the document to be used with
        this library.

        :param str password: The password to match.
        :return: ``0`` if the password failed, ``1`` if the password matched the user
            password, and ``2`` if the password matched the owner password.
        :rtype: int
        :raises NotImplementedError: if document uses an unsupported encryption
            method.
        """

        self._override_encryption = True
        try:
            return self._decrypt(password)
        finally:
            self._override_encryption = False

    def decode_permissions(self, permissions_code):
        # Takes the permissions as an integer, returns the allowed access
        permissions = {}
        permissions["print"] = permissions_code & (1 << 3 - 1) != 0  # bit 3
        permissions["modify"] = permissions_code & (1 << 4 - 1) != 0  # bit 4
        permissions["copy"] = permissions_code & (1 << 5 - 1) != 0  # bit 5
        permissions["annotations"] = permissions_code & (1 << 6 - 1) != 0  # bit 6
        permissions["forms"] = permissions_code & (1 << 9 - 1) != 0  # bit 9
        permissions["accessability"] = permissions_code & (1 << 10 - 1) != 0  # bit 10
        permissions["assemble"] = permissions_code & (1 << 11 - 1) != 0  # bit 11
        permissions["print_high_quality"] = (
            permissions_code & (1 << 12 - 1) != 0
        )  # bit 12
        return permissions

    def _decrypt(self, password):
        # Decrypts data as per Section 3.5 (page 117) of PDF spec v1.7
        # "The security handler defines the use of encryption and decryption in
        # the document, using the rules specified by the CF, StmF, and StrF entries"
        encrypt = self.trailer[TK.ENCRYPT].get_object()
        # /Encrypt Keys:
        # Filter (name)   : "name of the preferred security handler "
        # V (number)      : Algorithm Code
        # Length (integer): Length of encryption key, in bits
        # CF (dictionary) : Crypt filter
        # StmF (name)     : Name of the crypt filter that is used by default when decrypting streams
        # StrF (name)     : The name of the crypt filter that is used when decrypting all strings in the document
        # R (number)      : Standard security handler revision number
        # U (string)      : A 32-byte string, based on the user password
        # P (integer)     : Permissions allowed with user access
        if encrypt["/Filter"] != "/Standard":
            raise NotImplementedError(
                "only Standard PDF encryption handler is available"
            )
        if not (encrypt["/V"] in (1, 2)):
            raise NotImplementedError(
                "only algorithm code 1 and 2 are supported. This PDF uses code %s"
                % encrypt["/V"]
            )
        user_password, key = self._authenticate_user_password(password)
        if user_password:
            self._decryption_key = key
            return 1
        else:
            rev = encrypt["/R"].get_object()
            if rev == 2:
                keylen = 5
            else:
                keylen = encrypt[SA.LENGTH].get_object() // 8
            key = _alg33_1(password, rev, keylen)
            real_O = encrypt["/O"].get_object()
            if rev == 2:
                userpass = RC4_encrypt(key, real_O)
            else:
                val = real_O
                for i in range(19, -1, -1):
                    new_key = b_("")
                    for l in range(len(key)):
                        new_key += b_(chr(_utils.ord_(key[l]) ^ i))
                    val = RC4_encrypt(new_key, val)
                userpass = val
            owner_password, key = self._authenticate_user_password(userpass)
            if owner_password:
                self._decryption_key = key
                return 2
        return 0

    def _authenticate_user_password(self, password):
        encrypt = self.trailer[TK.ENCRYPT].get_object()
        rev = encrypt[ED.R].get_object()
        owner_entry = encrypt[ED.O].get_object()
        p_entry = encrypt[ED.P].get_object()
        if TK.ID in self.trailer:
            id_entry = self.trailer[TK.ID].get_object()
        else:
            # Some documents may not have a /ID, use two empty
            # byte strings instead. Solves
            # https://github.com/mstamy2/PyPDF2/issues/608
            id_entry = ArrayObject([ByteStringObject(b""), ByteStringObject(b"")])
        id1_entry = id_entry[0].get_object()
        real_U = encrypt[ED.U].get_object().original_bytes
        if rev == 2:
            U, key = _alg34(password, owner_entry, p_entry, id1_entry)
        elif rev >= 3:
            U, key = _alg35(
                password,
                rev,
                encrypt[SA.LENGTH].get_object() // 8,
                owner_entry,
                p_entry,
                id1_entry,
                encrypt.get(ED.ENCRYPT_METADATA, BooleanObject(False)).get_object(),
            )
            U, real_U = U[:16], real_U[:16]
        return U == real_U, key

    @property
    def is_encrypted(self):
        """
        Read-only boolean property showing whether this PDF file is encrypted.
        Note that this property, if true, will remain true even after the
        :meth:`decrypt()<PdfReader.decrypt>` method is called.
        """
        return TK.ENCRYPT in self.trailer

    def getIsEncrypted(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`is_encrypted` instead.
        """
        warnings.warn(
            "getIsEncrypted() will be removed in PyPDF2 2.0.0. "
            "Use the is_encrypted property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.is_encrypted

    @property
    def isEncrypted(self):
        """
        .. deprecated:: 1.28.0

            Use :py:attr:`is_encrypted` instead.
        """
        warnings.warn(
            "isEncrypted will be removed in PyPDF2 2.0.0. "
            "Use the is_encrypted property instead.",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.is_encrypted


class PdfFileReader(PdfReader):
    def __init__(self, *args, **kwargs):
        warnings.warn(
            "PdfFileReader was renamed to PdfReader. PdfFileReader will be removed",
            PendingDeprecationWarning,
            stacklevel=2,
        )
        if "strict" not in kwargs and len(args) < 2:
            kwargs["strict"] = True  # maintain the default
        super().__init__(*args, **kwargs)



import os

import PyPDF2
from PyPDF2 import PdfReader
from PyPDF2.generic import Destination

TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
PROJECT_ROOT = os.path.dirname(TESTS_ROOT)
RESOURCE_ROOT = os.path.join(PROJECT_ROOT, "Resources")


def page_ops(pdf_path, password):
    pdf_path = os.path.join(RESOURCE_ROOT, pdf_path)

    reader = PdfReader(pdf_path)

    if password:
        reader.decrypt(password)

    page = reader.pages[0]
    page.mergeRotatedScaledPage(page, 90, 1, 1)
    page.mergeScaledTranslatedPage(page, 1, 1, 1)
    page.mergeRotatedScaledTranslatedPage(page, 90, 1, 1, 1, 1)
    page.add_transformation([1, 0, 0, 0, 0, 0])
    page.scale(2, 2)
    page.scaleBy(0.5)
    page.scaleTo(100, 100)
    page.compress_content_streams()
    page.extract_text()


def test_page_operations(benchmark):
    """
    Apply various page operations.

    Rotation, scaling, translation, content stream compression, text extraction
    """
    benchmark(page_ops, "libreoffice-writer-password.pdf", "openpassword")


def merge():
    pdf_path = os.path.join(RESOURCE_ROOT, "crazyones.pdf")
    outline = os.path.join(RESOURCE_ROOT, "pdflatex-outline.pdf")
    pdf_forms = os.path.join(RESOURCE_ROOT, "pdflatex-forms.pdf")
    pdf_pw = os.path.join(RESOURCE_ROOT, "libreoffice-writer-password.pdf")

    merger = PyPDF2.PdfMerger()

    # string path:
    merger.append(pdf_path)
    merger.append(outline)
    merger.append(pdf_path, pages=PyPDF2.pagerange.PageRange(slice(0, 0)))
    merger.append(pdf_forms)

    # Merging an encrypted file
    reader = PyPDF2.PdfReader(pdf_pw)
    reader.decrypt("openpassword")
    merger.append(reader)

    # PdfReader object:
    merger.append(PyPDF2.PdfReader(pdf_path, "rb"), bookmark=True)

    # File handle
    with open(pdf_path, "rb") as fh:
        merger.append(fh)

    bookmark = merger.add_bookmark("A bookmark", 0)
    merger.add_bookmark("deeper", 0, parent=bookmark)
    merger.add_metadata({"author": "Martin Thoma"})
    merger.add_named_destionation("title", 0)
    merger.set_page_layout("/SinglePage")
    merger.set_page_mode("/UseThumbs")

    tmp_path = "dont_commit_merged.pdf"
    merger.write(tmp_path)
    merger.close()

    # Check if bookmarks are correct
    reader = PyPDF2.PdfReader(tmp_path)
    assert [
        el.title for el in reader._get_outlines() if isinstance(el, Destination)
    ] == [
        "A bookmark",
        "Foo",
        "Bar",
        "Baz",
        "Foo",
        "Bar",
        "Baz",
        "Foo",
        "Bar",
        "Baz",
        "True",
    ]

    # Clean up
    os.remove(tmp_path)


def test_merge(benchmark):
    """
    Apply various page operations.

    Rotation, scaling, translation, content stream compression, text extraction
    """
    benchmark(merge)
